import numpy as np
import pyvista as pv

from ntrfc.geometry.line import polyline_from_points, refine_spline
from ntrfc.math.vectorcalc import vecAngle
from ntrfc.turbo.cascade_geometry import calcmidpassagestreamline


def create_profileprobes(ssPoly, psPoly, midspan_z, pden_ps, pden_ss, tolerance=1e-10):
    """
    Create profile probes from two PolyData objects.

    Parameters:
    - ssPoly: PyVista PolyData object representing the suction side profile.
    - psPoly: PyVista PolyData object representing the pressure side profile.
    - midspan_z: Height of the midspan plane along the z-axis.
    - pden_ps: Density of the pressure side profile points.
    - pden_ss: Density of the suction side profile points.
    - tolerance: Small tolerance value to shift the 3D faces along their normals.

    Returns:
    - probes_ss: PyVista PolyData object representing the profile probes on the suction side.
    - probes_ps: PyVista PolyData object representing the profile probes on the pressure side.
    """

    # Refine the splines defined by the input PolyData objects
    ref_ss_x, ref_ss_y = refine_spline(ssPoly.points[::, 0], ssPoly.points[::, 1], 4000)
    ref_ps_x, ref_ps_y = refine_spline(psPoly.points[::, 0], psPoly.points[::, 1], 4000)

    # Create PolyData objects from the refined splines
    ref_ss_points = np.stack((ref_ss_x, ref_ss_y, np.zeros(len(ref_ss_y)))).T
    ref_ps_points = np.stack((ref_ps_x, ref_ps_y, np.zeros(len(ref_ps_y)))).T
    ref_ssPoly = pv.PolyData(ref_ss_points)
    ref_psPoly = pv.PolyData(ref_ps_points)

    # Convert the PolyData objects to polylines
    ref_ss_poly = polyline_from_points(ref_ssPoly.points)
    ref_ps_poly = polyline_from_points(ref_psPoly.points)

    # Extrude the polylines along the z-axis to create 3D face models
    ref_ss_face = ref_ss_poly.extrude((0, 0, midspan_z * 2)).compute_normals()
    ref_ps_face = ref_ps_poly.extrude((0, 0, midspan_z * 2)).compute_normals()

    # Shift the 3D faces slightly along their normals
    ref_ss_face_shift = ref_ss_face.copy()
    ref_ss_face_shift.points += tolerance * ref_ss_face_shift.point_data["Normals"]
    ref_ps_face_shift = ref_ps_face.copy()
    ref_ps_face_shift.points += tolerance * ref_ps_face_shift.point_data["Normals"]

    # Create a cut through each 3D face at the midspan plane
    ref_ss_cut = ref_ss_face_shift.slice(normal="z", origin=(0, 0, midspan_z))
    ref_ps_cut = ref_ps_face_shift.slice(normal="z", origin=(0, 0, midspan_z))

    # Extract the x and y coordinates of the points on the cut faces
    x_ss_shift = ref_ss_cut.points[::, 0]
    y_ss_shift = ref_ss_cut.points[::, 1]
    x_ps_shift = ref_ps_cut.points[::, 0]
    y_ps_shift = ref_ps_cut.points[::, 1]

    # Refine the splines defined by the cut faces using the specified densities
    x_bl_ss, y_bl_ss = refine_spline(x_ss_shift, y_ss_shift, pden_ss)
    x_bl_ps, y_bl_ps = refine_spline(x_ps_shift, y_ps_shift, pden_ps)

    # Create PolyData objects from the refined splines
    probes_ps = pv.PolyData(np.stack((x_bl_ss, y_bl_ss, midspan_z * np.ones(len(x_bl_ss)))).T)
    probes_ss = pv.PolyData(np.stack((x_bl_ps, y_bl_ps, midspan_z * np.ones(len(x_bl_ps)))).T)

    # Return the profile probes
    return probes_ss, probes_ps


def create_midpassageprobes(midspan_z, x_inlet, x_outlet, pitch, beta1, beta2, midspoly, nop_streamline):
    x_mcl, y_mcl = midspoly.points[::, 0], midspoly.points[::, 1]

    x_mpsl, y_mpsl = calcmidpassagestreamline(x_mcl, y_mcl, beta1, beta2, x_inlet, x_outlet, pitch)

    # x_probes = []
    # y_probes = []
    z_probes = []

    nop = int(nop_streamline)

    xn, yn = refine_spline(x_mpsl, y_mpsl, nop)
    for i in range(nop):
        z_probes.append(midspan_z)

    x_probes = xn
    y_probes = yn

    dist = np.sqrt((x_probes[0] - x_probes[1]) ** 2 + (y_probes[0] - y_probes[1]) ** 2)

    x_probes[0] = x_probes[0] + 0.00001 * dist
    x_probes[-1] = x_probes[-1] - 0.00001 * dist

    return pv.PolyData(np.stack([x_probes, y_probes, np.ones(len(x_probes) * midspan_z)]).T)


def create_stagnationpointprobes(length, nop, sortedpoly, ind_vk, u_inlet, midspan_z):
    vk_point = sortedpoly.points[ind_vk]
    angle = vecAngle(u_inlet, np.array([1, 0, 0]) * 180 / np.pi)
    stagnationLine = pv.Line((0, 0, 0), (-length, 0, 0), nop - 1)
    stagnationLine.rotate_z(angle, inplace=False)
    stagnationLine.translate(vk_point, inplace=False)
    x_probes = stagnationLine.points[::, 0]
    y_probes = stagnationLine.points[::, 1]
    z_probes = stagnationLine.points[::, 2] + midspan_z
    return pv.PolyData(np.stack([x_probes, y_probes, z_probes]).T)



import os
import tempfile
from dataclasses import dataclass

import pyvista as pv

from ntrfc.math.vectorcalc import vecAbs
from ntrfc.turbo.pointcloud_methods import extract_geo_paras


@dataclass
class DomainParameters:
    """
    A class representing the geometrical parameters of a simulation domain
    for a turbomachinery linear cascade simulation.

    Attributes:
        stagger_angle (float): The stagger angle of the blades.
        sspoly (pv.PolyData): A pv.PolyData object representing the suction side poly.
        pspoly (pv.PolyData): A pv.PolyData object representing the pressure side poly.
        xinlet (float): The x coordinate of the inlet.
        xoutlet (float): The x coordinate of the outlet.
        pitch (float): The pitch of the blades.
        blade_yshift (float): The y shift of the blades.
        beta_in (float): The metal angle of the profile at the inlet.
        beta_out (float): The metal angle of the profile at the outlet.
        alpha (float): The alpha shape of the blade.
        profile_points (pv.PolyData): A pv.PolyData object representing the profile points.
        leading_edge_index (int): The index of the leading edge point in the profile_points array.
        trailing_edge_index (int): The index of the trailing edge point in the profile_points array.
    """
    stagger_angle: float = None
    sspoly: pv.PolyData = None
    pspoly: pv.PolyData = None
    midspoly: pv.PolyData = None
    xinlet: float = None
    xoutlet: float = None
    pitch: float = None
    blade_yshift: float = None
    beta_in: float = None
    beta_out: float = None
    alpha: float = None
    profile_points: pv.PolyData = None
    leading_edge_index: int = None
    trailing_edge_index: int = None
    chordlength: float = None
    xinlet: int = None
    xoutlet: int = None

    def plot_domainparas(self, figurepath=tempfile.mkdtemp() + "/plot.png"):
        """
        Plot the domain parameters using PyVista.


        Returns:
            pv.Plotter: The PyVista plotter object used for plotting.
        """
        if os.getenv('DISPLAY') is None:
            pv.start_xvfb()  # Start X virtual framebuffer (Xvfb)
        pv.set_plot_theme('document')
        plotter = pv.Plotter(off_screen=True)
        plotter.window_size = 1600, 1600
        # Plot the suction side and pressure side polys
        plotter.add_mesh(self.sspoly, color='b', show_edges=True, label="suction-side")
        plotter.add_mesh(self.pspoly, color='r', show_edges=True, label="pressure-side")
        plotter.add_mesh(self.midspoly, color='k', label="camber")

        # Plot the profile points
        plotter.add_points(self.profile_points, point_size=1, color='w', opacity=0.8)
        plotter.add_legend()
        plotter.add_axes()
        plotter.view_xy()
        plotter.screenshot(figurepath)
        return figurepath

    def generate_params_by_pointcloud(self, points, alpha=None):
        sortedPoly, psPoly, ssPoly, ind_vk, ind_hk, midsPoly, beta_leading, beta_trailing, camber_angle, alpha = extract_geo_paras(
            points, alpha)

        self.profile_points = sortedPoly
        self.sspoly = ssPoly
        self.pspoly = psPoly
        self.midspoly = midsPoly
        self.leading_edge_index = ind_vk
        self.trailing_edge_index = ind_hk
        self.beta_in = beta_leading
        self.beta_out = beta_trailing
        self.stagger_angle = camber_angle
        self.alpha = alpha
        self.chordlength = vecAbs(
            self.profile_points.points[self.leading_edge_index] - self.profile_points.points[self.trailing_edge_index])
        self.xinlet = -self.chordlength + self.profile_points.points[self.leading_edge_index][0]
        self.xoutlet = self.profile_points.points[self.trailing_edge_index][0] + self.chordlength



import numpy as np
import pyvista as pv
from scipy.interpolate import interp1d


def polyline_from_points(points):
    poly = pv.PolyData()
    poly.points = points
    the_cell = np.arange(0, len(points), dtype=np.int_)
    the_cell = np.insert(the_cell, 0, len(points))
    poly.lines = the_cell
    return poly


def lines_from_points(points):
    """Given an array of points, make a line set"""
    poly = pv.PolyData()
    poly.points = points
    cells = np.full((len(points) - 1, 3), 2, dtype=np.int_)
    cells[:, 1] = np.arange(0, len(points) - 1, dtype=np.int_)
    cells[:, 2] = np.arange(1, len(points), dtype=np.int_)
    poly.lines = cells
    return poly


def refine_spline(x, y, res, half_cosine_spacing=False):
    """
    https://stackoverflow.com/questions/51512197/python-equidistant-points-along-a-line-joining-set-of-points/51515357

    :param x:
    :param y:
    :param res:
    :return: x,y
    """

    distance = np.cumsum(np.sqrt(np.ediff1d(x, to_begin=0) ** 2 + np.ediff1d(y, to_begin=0) ** 2))
    distance = distance / distance[-1]

    fx, fy = interp1d(distance, x), interp1d(distance, y)

    if half_cosine_spacing:
        beta = np.linspace(0.0, np.pi, res)
        alpha_ = [(0.5 * (1.0 - np.cos(xa))) for xa in beta]  # Half cosine based spacing
    else:
        alpha_ = np.linspace(0, 1, res)
    x, y = fx(alpha_), fy(alpha_)
    return x, y



from ntrfc.turbo.bladeloading import calc_cf


def test_calc_inflow_cp():
    from ntrfc.turbo.bladeloading import calc_inflow_cp
    # Test case with positive pressure difference
    px = 101325  # Pa
    pt1 = 110000  # Pa
    p1 = 100000  # Pa
    expected_cp = 0.8675
    assert calc_inflow_cp(px, pt1, p1) == expected_cp

    # Test case with negative pressure difference
    px = 90000  # Pa
    pt1 = 110000  # Pa
    p1 = 100000  # Pa
    expected_cp = 2
    assert calc_inflow_cp(px, pt1, p1) == expected_cp

    # Test case with zero pressure difference
    px = 100000  # Pa
    pt1 = 100000  # Pa
    p1 = 110000  # Pa
    expected_cp = 0
    assert calc_inflow_cp(px, pt1, p1) == expected_cp


def test_calc_zeta():
    from ntrfc.turbo.bladeloading import calc_zeta
    assert calc_zeta(100, 80, 70) == 0.6666666666666666
    assert calc_zeta(50, 20, 10) == 0.75


def test_calc_cf():
    import numpy as np
    # Test case 1
    tau_w = 0.05
    u_inf = 10
    rho_inf = 1.225
    expected_cf = 0.0008163265306122448
    actual_cf = calc_cf(tau_w, u_inf, rho_inf)
    assert np.isclose(actual_cf, expected_cf)

    # Test case 2
    tau_w = 0.1
    u_inf = 20
    rho_inf = 1.5
    expected_cf = 0.0003333333333333334
    actual_cf = calc_cf(tau_w, u_inf, rho_inf)
    assert np.isclose(actual_cf, expected_cf)



import numpy as np
from scipy.integrate import simps
from scipy.signal import argrelmax

from ntrfc.math.methods import autocorr, zero_crossings


def integralscales(signal, timesteparray):
    mean = np.mean(signal)
    fluctations = signal - mean
    timesteps = timesteparray.copy()
    autocorrelated = autocorr(fluctations)
    # we are integrating from zero to zero-crossing in the autocorrelation, we need the time to begin with zeros
    # probably the used datasample is not beginning with 0. therefore:
    timesteps -= timesteps[0]
    if len(zero_crossings(autocorrelated)) > 0:
        acorr_zero_crossings = zero_crossings(autocorrelated)[0]
    else:
        print("no zero crossing found, using first minimal value (possibly last timestep). check data quality!")
        acorr_zero_crossings = np.where(autocorrelated == min(autocorrelated))[0][0]

    if all(np.isnan(autocorrelated)) or acorr_zero_crossings == 0:
        return 0
    integral_time_scale = simps(autocorrelated[:acorr_zero_crossings], timesteps[:acorr_zero_crossings])
    integral_length_scale = integral_time_scale * mean

    return integral_time_scale, integral_length_scale


def get_self_correlating_frequencies(timesteps, signal):
    """
    Calculate the self-correlating frequencies in a signal.

    Parameters
    ----------
    timesteps : ndarray
        An array of uniformly spaced timesteps, corresponding to the time intervals between samples in the signal.
    signal : ndarray
        The input signal to analyze.

    Returns
    -------
    frequencies : ndarray
        An array of self-correlating frequencies, in units of inverse timesteps.

    Notes
    -----
    This function calculates the autocorrelation of the input signal, and finds the local maxima in the autocorrelation signal. These local maxima correspond to the self-correlating frequencies in the signal.
    """
    # Calculate the autocorrelation of the signal
    autocorr_signal = autocorr(signal)

    # Find self-correlating frequencies by finding the local maxima in the autocorrelation signal
    self_correlating_frequencies = argrelmax(autocorr_signal)[0]
    ts_shifted = timesteps - timesteps[0]
    periods = ts_shifted[self_correlating_frequencies]
    frequencies = 1 / periods
    return frequencies



from ntrfc.geometry.plane import inside_poly


def test_extract_vk_hk(verbose=False):
    """
    tests a NACA  profile in a random angle as a minimal example.
    :return:
    """
    from ntrfc.turbo.profile_tele_extraction import extract_vk_hk
    from ntrfc.turbo.airfoil_generators.naca_airfoil_creator import naca
    import numpy as np
    import pyvista as pv

    # d1,d2,d3,d4 = np.random.randint(0,9),np.random.randint(0,9),np.random.randint(0,9),np.random.randint(0,9)
    # digitstring = str(d1)+str(d2)+str(d3)+str(d4)
    # manifold problems with other profiles with veronoi-mid and other unoptimized code. therefor tests only 0009
    # todo: currently we cant test half_cosine_spacing profiles, as the resolution is too good for extract_vk_hk
    naca_code = "6509"
    angle = 30  # deg
    res = 400
    xs, ys = naca(naca_code, res, half_cosine_spacing=False, finite_te=False)
    sorted_poly = pv.PolyData(np.stack([xs[:-1], ys[:-1], np.zeros(len(xs) - 1)]).T)
    sorted_poly.rotate_z(angle, inplace=True)
    X, Y = sorted_poly.points[::, 0], sorted_poly.points[::, 1]
    ind_1 = res
    ind_2 = 0

    points = np.stack((X[:], Y[:], np.zeros(len(X)))).T
    profile_points = pv.PolyData(points)
    sorted_poly = pv.PolyData(profile_points)
    ind_vk, ind_hk = extract_vk_hk(sorted_poly, verbose=verbose)

    if verbose:
        p = pv.Plotter()
        p.add_mesh(sorted_poly.points[ind_hk], color="yellow", point_size=20)
        p.add_mesh(sorted_poly.points[ind_vk], color="red", point_size=20)
        p.add_mesh(sorted_poly)
        p.show()

    assert ind_hk == ind_2, "wrong hk-index chosen"
    assert ind_vk == ind_1, "wrong vk-index chosen"


def test_midline_from_sides(verbose=False):
    from ntrfc.turbo.pointcloud_methods import midline_from_sides
    from ntrfc.math.vectorcalc import vecAbs
    from ntrfc.turbo.airfoil_generators.naca_airfoil_creator import naca
    from ntrfc.turbo.pointcloud_methods import extractSidePolys
    import numpy as np
    import pyvista as pv

    res = 240
    x, y = naca('0009', res, half_cosine_spacing=True)
    ind_hk = 0
    ind_vk = res

    points = np.stack((x[:], y[:], np.zeros(res * 2))).T
    poly = pv.PolyData(points)
    sspoly, pspoly = extractSidePolys(ind_hk, ind_vk, poly)

    mids = midline_from_sides(pspoly, sspoly)

    length = mids.length
    testlength = vecAbs(sspoly.points[0] - sspoly.points[-1])

    if verbose:
        poly["ids"] = np.arange(poly.number_of_points)
        p = pv.Plotter()
        p.add_mesh(poly)
        p.add_mesh(mids)
        p.add_mesh(poly.points[ind_hk], color="k")
        p.add_mesh(poly.points[ind_vk], color="k")
        p.show()

    assert length == testlength, "midline not accurate"


def test_midLength():
    """
    checks weather
    """
    from ntrfc.turbo.pointcloud_methods import mid_length
    import numpy as np
    import pyvista as pv

    radius = 0.5
    res = 100
    mid = int(res / 2)
    theta = np.linspace(0, 2 * np.pi, 100)

    x = radius * np.cos(theta)
    y = radius * np.sin(theta)

    fake_vk = 0
    fake_hk = mid
    circle = pv.PolyData(np.stack([x, y, np.zeros(len(x))]).T)
    length = mid_length(fake_vk, fake_hk, circle)
    assert np.isclose(2 * radius, length, rtol=1e-4), "length should be two times the size of the defined test-circle"


def test_extract_geo_paras(verbose=False):
    from ntrfc.turbo.pointcloud_methods import extract_geo_paras
    from ntrfc.turbo.airfoil_generators.naca_airfoil_creator import naca
    import numpy as np
    import pyvista as pv

    naca_code = "6009"
    angle = 10  # deg
    alpha = 1
    res = 240
    xs, ys = naca(naca_code, res, half_cosine_spacing=True, finite_te=False)
    sorted_poly = pv.PolyData(np.stack([xs, ys, np.zeros(len(xs))]).T)
    sorted_poly.rotate_z(angle, inplace=True)

    poly, ps_poly, ss_poly, ind_hk, ind_vk, mids_poly, beta_leading, beta_trailing, camber_angle, alpha = extract_geo_paras(
        sorted_poly, alpha)

    if verbose:
        p = pv.Plotter()
        p.add_mesh(ss_poly, color="g", label="ssPoly")
        p.add_mesh(ps_poly, color="b", label="psPoly")
        p.add_mesh(mids_poly)
        p.add_mesh(poly.points[ind_hk], color="w", label="ind_hk")
        p.add_mesh(poly.points[ind_vk], color="k", label="ind_vk")
        p.add_legend()
        p.show()

    assert np.isclose(beta_leading, angle, rtol=0.03), "wrong leading edge angle"
    assert np.isclose(beta_trailing, angle, rtol=0.03), "wrong leading edge angle"
    assert np.isclose(mids_poly.length, 1, rtol=0.03)
    assert np.isclose(camber_angle, angle, rtol=0.03)
    assert alpha > 0


def test_inside_poly():
    # Test for a simple polygon and point
    polygon = [(0, 0), (0, 1), (1, 1), (1, 0)]
    point = [(0.5, 0.5)]
    assert inside_poly(polygon, point)[0] == True

    # Test for a point outside the polygon
    polygon = [(0, 0), (0, 1), (1, 1), (1, 0)]
    point = [(1.5, 1.5)]
    assert inside_poly(polygon, point)[0] == False

    # Test for a point on the boundary of the polygon
    polygon = [(0, 0), (0, 1), (1, 1), (1, 0)]
    point = [(1, 1)]
    assert inside_poly(polygon, point)[0] == False



import numpy as np

from ntrfc.geometry.line import refine_spline


def calcmidpassagestreamline(x_mcl, y_mcl, beta1, beta2, x_inlet, x_outlet, t):
    """
    Returns mid-passage line from sceletal-line
    Returns two lists of Points representing a curve through the passage
    Input:
    x_mcl = Tuple
    y_mcl = Tuple
    beta1, beta2 = Angle in deg - Beta = Anströmwinkel
    x_inlet, x_outlet = scalar - representing position x-component of in/outlet
    t = scalar pitch
    """

    delta_x_vk = x_mcl[0] - x_inlet
    delta_y_vk = np.tan(np.deg2rad(beta1 - 90)) * delta_x_vk

    p_inlet_x = x_mcl[0] - delta_x_vk
    p_inlet_y = y_mcl[0] - delta_y_vk

    delta_x_hk = x_outlet - x_mcl[-1]
    delta_y_hk = delta_x_hk * np.tan(np.deg2rad(beta2 - 90))

    p_outlet_x = x_mcl[-1] + delta_x_hk
    p_outlet_y = y_mcl[-1] + delta_y_hk

    x_mpsl = [p_inlet_x] + list(x_mcl) + [p_outlet_x]
    y_mpsl = [p_inlet_y] + list(y_mcl) + [p_outlet_y]

    for i in range(len(x_mpsl)):
        y_mpsl[i] = y_mpsl[i] + 0.5 * t

    return refine_spline(x_mpsl, y_mpsl, 1000)



import numpy as np
import pytest

from ntrfc.cascade_case.utils.probecontainer import ProbeContainer


@pytest.fixture
def probe_container():
    container = ProbeContainer()
    container.add_probe(
        position=np.array([0, 0, 0]),
        values={'value1': np.array([1, 2, 3]), 'value2': np.array([[1, 2, 3], [4, 5, 6], [3, 2, 1]])},
        groupname='Group1',
        timesteps=np.array([0, 1, 2])
    )
    container.add_probe(
        position=np.array([1, 1, 1]),
        values={'value2': np.array([[1, 2, 3], [4, 2, 4], [7, 8, 9]]), 'value1': np.array([4, 5, 6])},
        groupname='Group2',
        timesteps=np.array([0, 1, 2])
    )
    container.add_probe(
        position=np.array([0, 0, 0]),
        values={'value2': np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]]), 'value1': np.array([2, 2, 2])},
        groupname='Group2',
        timesteps=np.array([0, 1, 2])
    )
    return container


def test_add_probe(probe_container):
    assert len(probe_container.probes) == 3


def test_filter_probes_by_groupname(probe_container):
    filtered_probes = probe_container.filter_probes(by_groupname='Group1')
    assert len(filtered_probes) == 1
    assert filtered_probes[0].groupname == 'Group1'


def test_filter_probes_by_position(probe_container):
    filtered_probes = probe_container.filter_probes(by_position=np.array([0, 0, 0]))
    assert len(filtered_probes) == 2
    assert np.array_equal(filtered_probes[0].position, np.array([0, 0, 0]))
    assert np.array_equal(filtered_probes[1].position, np.array([0, 0, 0]))


def test_plot_probes(probe_container):
    with pytest.raises(AssertionError):
        probe_container.plot_probes()  # value not specified

    # Test plotting with value='value1'
    probe_container.plot_probes(value='value1')

    # Test plotting with value='value2'
    probe_container.plot_probes(value='value2')

    # Add more specific tests based on your requirements



def test_naca():
    from ntrfc.turbo.airfoil_generators.naca_airfoil_creator import naca
    import numpy as np
    import pyvista as pv
    res = 480

    xs, ys = naca("6510", res, half_cosine_spacing=False, finite_te=False)
    points = np.stack([xs, ys, np.zeros(len(xs))]).T
    poly = pv.PolyData(points)
    assert poly.number_of_points == res * 2

    xs, ys = naca("6510", res, half_cosine_spacing=True, finite_te=False)
    points = np.stack([xs, ys, np.zeros(len(xs))]).T
    poly = pv.PolyData(points)
    assert poly.number_of_points == res * 2

    xs, ys = naca("6510", res, half_cosine_spacing=False, finite_te=True)
    points = np.stack([xs, ys, np.zeros(len(xs))]).T
    poly = pv.PolyData(points)
    assert poly.number_of_points == res * 2 - 1 + 100

    xs, ys = naca("6510", res, half_cosine_spacing=True, finite_te=True)
    points = np.stack([xs, ys, np.zeros(len(xs))]).T
    poly = pv.PolyData(points)
    assert poly.number_of_points == res * 2 - 1 + 100

    xs, ys = naca("23112", res, half_cosine_spacing=False, finite_te=False)
    points = np.stack([xs, ys, np.zeros(len(xs))]).T
    poly = pv.PolyData(points)
    assert poly.number_of_points == res * 2

    xs, ys = naca("23112", res, half_cosine_spacing=True, finite_te=False)
    points = np.stack([xs, ys, np.zeros(len(xs))]).T
    poly = pv.PolyData(points)
    assert poly.number_of_points == res * 2

    xs, ys = naca("23112", res, half_cosine_spacing=False, finite_te=True)
    points = np.stack([xs, ys, np.zeros(len(xs))]).T
    poly = pv.PolyData(points)
    assert poly.number_of_points == res * 2 - 1 + 100

    xs, ys = naca("23112", res, half_cosine_spacing=True, finite_te=True)
    points = np.stack([xs, ys, np.zeros(len(xs))]).T
    poly = pv.PolyData(points)
    assert poly.number_of_points == res * 2 - 1 + 100



import os

import pyvista as pv

ON_CI = 'CI' in os.environ

if ON_CI:
    pv.start_xvfb()


def test_optimal_timewindow(verbose=False):
    from ntrfc.timeseries.stationarity import optimal_window_size
    import numpy as np
    # sine
    res = 10000

    nper = 8
    x = np.linspace(0, 2 * np.pi * nper, res)
    min_interval = 0.05
    max_interval = 0.25
    # sin
    # we have four periods, at least one period should be captured
    # thats res // 4 as a return
    ysin = np.sin(x)
    opt_window, opt_window_size, nperiods = optimal_window_size(ysin, min_interval, max_interval)
    assert opt_window_size == res // nper * nperiods
    assert nperiods == 2
    # tanh
    tanh = np.tanh(x * 2)
    opt_window, opt_window_size, nperiods = optimal_window_size(tanh, min_interval, max_interval)
    assert opt_window_size == res * min_interval
    assert nperiods == 0
    # euler
    eul = np.e ** (-x * 60)
    opt_window, opt_window_size, nperiods = optimal_window_size(eul, min_interval, max_interval)
    assert opt_window_size == res * min_interval
    assert nperiods == 0


def test_stationarity_uncertainties_stationarysine(verbose=False):
    from ntrfc.timeseries.stationarity import estimate_stationarity
    from ntrfc.math.methods import reldiff
    import numpy as np
    from itertools import product
    import matplotlib.pyplot as plt

    def signalgen_sine(amplitude, frequency, mean, time):
        frequency_resolution = 256
        tau = frequency ** -1

        step = tau / frequency_resolution
        times = np.arange(0, time, step)
        values = amplitude * np.sin(frequency * (2 * np.pi) * times) + mean
        return times, values

    test_amplitudes = [1]  # 1,10
    test_frequencies = [1]  # 1,10
    test_times = [20]  # 10, 60
    test_mean = [-1]  # -1,0,1

    maxperiods = -1
    minperiods = np.inf

    test_configs = list(product(test_amplitudes, test_frequencies, test_times, test_mean))

    for amplitude, frequency, time, mean in test_configs:
        equals_periods = time / frequency ** -1
        if equals_periods > maxperiods:
            maxperiods = int(equals_periods)
        if minperiods > equals_periods:
            minperiods = int(equals_periods)
        timesteps, values = signalgen_sine(amplitude=amplitude, frequency=frequency, mean=mean, time=time)
        stationary_timestep = estimate_stationarity(values)
        analytic_stationary_limit = 0.0
        if verbose:
            plt.figure()
            plt.plot(timesteps, values)
            plt.axvline(stationary_timestep)
            plt.show()
        assert 0.05 > reldiff(time - analytic_stationary_limit,
                              time - timesteps[stationary_timestep]), "computation failed"


def test_stationarity_uncertainties_abatingsine(verbose=False):
    from ntrfc.timeseries.stationarity import estimate_stationarity
    from ntrfc.math.methods import reldiff
    import numpy as np
    from itertools import product
    import matplotlib.pyplot as plt

    def signalgen_abatingsine(amplitude, frequency, mean, abate, time):
        resolution = 64
        step = (1 / frequency) / resolution
        times = np.arange(0, time, step)
        values = amplitude * np.sin(frequency * (2 * np.pi) * times) + mean + np.e ** -(times * abate)
        return times, values

    test_amplitudes = [0.3]
    test_frequencies = [6]
    test_times = [20]
    test_mean = [-1]
    test_abate = [2, 5]

    test_configs = list(product(test_amplitudes, test_frequencies, test_times, test_mean, test_abate))

    for amplitude, frequency, time, mean, abate in test_configs:

        timesteps, values = signalgen_abatingsine(amplitude=amplitude, frequency=frequency, mean=mean, time=time,
                                                  abate=abate)
        stationary_timestep = estimate_stationarity(values)

        well_computed_stationarity_limit = -np.log(0.01) / abate
        well_computed_stationary_time = timesteps[-1] - well_computed_stationarity_limit
        stationary_time = timesteps[-1] - timesteps[stationary_timestep]
        if verbose:
            plt.figure()
            plt.plot(timesteps, values)
            plt.axvline(timesteps[stationary_timestep], color="green")
            plt.axvline(well_computed_stationarity_limit, color="red", label="computed")
            plt.legend()
            plt.show()
        assert 0.05 > reldiff(stationary_time, well_computed_stationary_time), "computation failed"


def test_stationarity_uncertainties_abatingsinenoise(verbose=False):
    from ntrfc.timeseries.stationarity import estimate_stationarity
    from ntrfc.math.methods import reldiff
    import numpy as np
    from itertools import product
    import matplotlib.pyplot as plt

    def signalgen_abatingsine(amplitude, noiseamplitude, frequency, mean, abate, time):
        resolution = 64
        step = (1 / frequency) / resolution

        times = np.arange(0, time, step)
        noise = np.random.normal(-1, 1, len(times)) * noiseamplitude

        values = amplitude * np.sin(frequency * (2 * np.pi) * times) + mean + np.e ** -(times * abate) + noise
        return times, values

    test_amplitudes = [0.1]
    test_noiseamplitude = [0.01]
    test_frequencies = [6]
    test_times = [40]
    test_mean = [-1]
    test_abate = [3]

    test_configs = list(
        product(test_amplitudes, test_noiseamplitude, test_frequencies, test_times, test_mean, test_abate))

    for amplitude, noiseamplitude, frequency, time, mean, abate in test_configs:

        timesteps, values = signalgen_abatingsine(amplitude=amplitude, noiseamplitude=noiseamplitude,
                                                  frequency=frequency, mean=mean, time=time,
                                                  abate=abate)
        stationary_timestep = estimate_stationarity(values)

        well_computed_stationarity_limit = -np.log(0.01) / abate
        well_computed_stationary_time = timesteps[-1] - well_computed_stationarity_limit
        stationary_time = timesteps[-1] - timesteps[stationary_timestep]
        if verbose:
            plt.figure()
            plt.plot(timesteps, values)
            plt.axvline(timesteps[stationary_timestep], color="green")
            plt.axvline(well_computed_stationarity_limit, color="red")
            plt.show()
        assert 0.05 >= reldiff(stationary_time, well_computed_stationary_time), "computation failed"


def test_stationarity_transientonly(verbose=False):
    from ntrfc.timeseries.stationarity import estimate_stationarity
    import numpy as np
    from itertools import product
    import matplotlib.pyplot as plt

    def signalgen_abatingsine(amplitude, frequency, mean, time):
        resolution = 36
        step = (1 / frequency) / resolution

        times = np.arange(0, time, step)

        values = (amplitude + frequency * (2 * np.pi) * times / 1600) * np.sin(frequency * (2 * np.pi) * times) + mean
        return times, values

    test_amplitudes = [0.1]
    test_frequencies = [6]
    test_times = [40]
    test_mean = [-2]

    test_configs = list(
        product(test_amplitudes, test_frequencies, test_times, test_mean))
    for amplitude, frequency, time, mean in test_configs:

        timesteps, values = signalgen_abatingsine(amplitude=amplitude,
                                                  frequency=frequency, mean=mean, time=time)

        statidx = estimate_stationarity(values)

        if verbose:
            plt.figure()
            plt.plot(timesteps, values)
            plt.show()
        assert statidx == False


def test_stationarity_transientonly():
    from ntrfc.timeseries.stationarity import estimate_stationarity
    import numpy as np

    res = 20000

    values = np.linspace(0, 10, res)
    stationary_timestep = estimate_stationarity(values)
    assert stationary_timestep == False


def test_stationarity_uncertainties_abating(verbose=False):
    from ntrfc.timeseries.stationarity import estimate_stationarity
    import numpy as np
    from itertools import product
    import matplotlib.pyplot as plt

    from ntrfc.math.methods import reldiff
    def signalgen_abating(noiseamplitude, mean, abate, time):
        resolution = 20000
        step = (time / resolution)

        times = np.arange(0, time, step)
        noise = np.random.normal(-1, 1, len(times)) * noiseamplitude

        values = mean + np.e ** -(times * abate) + noise
        return times, values

    test_noiseamplitude = [0.01]
    test_times = [30]
    test_mean = [-1]
    test_abate = [3, 2]

    test_configs = list(product(test_noiseamplitude, test_times, test_mean, test_abate))

    for noiseamplitude, time, mean, abate in test_configs:

        timesteps, values = signalgen_abating(noiseamplitude=noiseamplitude, mean=mean, abate=abate, time=time)
        stationary_timestep = estimate_stationarity(values)

        well_computed_stationarity_limit = -np.log(0.01) / abate
        well_computed_stationary_time = timesteps[-1] - well_computed_stationarity_limit
        stationary_time = timesteps[-1] - timesteps[stationary_timestep]
        if verbose:
            plt.figure()
            plt.plot(timesteps, values)
            plt.axvline(timesteps[stationary_timestep], color="green")
            plt.axvline(well_computed_stationarity_limit, color="red")
            plt.show()
        assert 0.05 >= reldiff(stationary_time, well_computed_stationary_time), "computation failed"



import tempfile
from dataclasses import dataclass, field

import matplotlib.pyplot as plt
import numpy as np


@dataclass
class Probe:
    position: np.array
    values: dict
    groupname: str
    timesteps: np.array


@dataclass
class ProbeContainer:
    # i need a default value for probes with an empty list
    probes: list = field(default_factory=lambda: [])

    # add function that appends a new probe to the list of probes

    def add_probe(self, position, values, groupname, timesteps):
        self.probes.append(Probe(position, values, groupname, timesteps))

    def filter_probes(self, by_groupname=None, by_position=None):
        filtered_probes = self.probes

        if by_groupname:
            filtered_probes = [probe for probe in filtered_probes if probe.groupname == by_groupname]

        if by_position is not None:
            filtered_probes = [probe for probe in filtered_probes if np.array_equal(probe.position, by_position)]

        return filtered_probes

    def plot_probes(self, groupname=None, positions=None, value=None, path=None):
        assert value, "a value must be specified"
        if path is None:
            tmpdir = tempfile.mkdtemp()
            path = tmpdir + "/plot.png"
        plotprobelist = self.filter_probes(groupname, positions)
        plt.figure(figsize=(32, 24), dpi=200)
        for probe in plotprobelist:
            if probe.values[value][0].shape == (1,):
                plt.plot(probe.timesteps, probe.values[value].T[0], label=f"{probe.position}")
            elif probe.values[value][0].shape == (3,):
                color = tuple(np.random.rand(3))
                plt.plot(probe.timesteps, probe.values[value][:, 0], label=f"{probe.position}", color=color)
                plt.plot(probe.timesteps, probe.values[value][:, 1], color=color)
                plt.plot(probe.timesteps, probe.values[value][:, 2], color=color)
        plt.legend()
        plt.savefig(path)



import tempfile

import numpy as np
import pyvista as pv

from ntrfc.turbo.pointcloud_methods import extract_geo_paras, calcMidPassageStreamLine


def cascade_2d_domain(profilepoints2d, x_inlet, x_outlet, pitch, unit, blade_shift, alpha, path=False):
    """
    profilepoints2d = 2d numpy array
    """

    if path is None:
        tmpdir = tempfile.mkdtemp()
        path = tmpdir + "/plot.png"
    # =============================================================================
    # Daten Einlesen
    # =============================================================================
    unitcoeff = 0
    if unit == "m":
        unitcoeff = 1
    elif unit == "mm":
        unitcoeff = 1 / 1000
    profilepoints2d.points *= unitcoeff

    # =============================================================================
    # Bestimmung Profilparameter
    # =============================================================================
    sortedPoly, psPoly, ssPoly, ind_vk, ind_hk, midsPoly, beta_leading, beta_trailing, camber_angle, alpha = extract_geo_paras(
        profilepoints2d,
        alpha,
        path)

    ##############################################################
    x_mids = midsPoly.points[::, 0]
    y_mids = midsPoly.points[::, 1]
    # x_ss = ssPoly.points[::, 0]
    # y_ss = ssPoly.points[::, 1]
    # x_ps = psPoly.points[::, 0]
    # y_ps = psPoly.points[::, 1]

    x_mpsl, y_mpsl = calcMidPassageStreamLine(x_mids, y_mids, beta_leading, beta_trailing,
                                              x_inlet, x_outlet, pitch)

    y_upper = np.array(y_mpsl) + blade_shift
    per_y_upper = pv.lines_from_points(np.stack((np.array(x_mpsl),
                                                 np.array(y_upper),
                                                 np.zeros(len(x_mpsl)))).T)
    y_lower = y_upper - pitch
    per_y_lower = pv.lines_from_points(np.stack((np.array(x_mpsl),
                                                 np.array(y_lower),
                                                 np.zeros(len(x_mpsl)))).T)

    inlet_pts = np.array([per_y_lower.points[per_y_lower.points[::, 0].argmin()],
                          per_y_upper.points[per_y_upper.points[::, 0].argmin()]])

    inletPoly = pv.Line(*inlet_pts)
    outlet_pts = np.array([per_y_lower.points[per_y_lower.points[::, 0].argmax()],
                           per_y_upper.points[per_y_upper.points[::, 0].argmax()]])

    outletPoly = pv.Line(*outlet_pts)

    p = pv.Plotter(off_screen=True)
    p.add_mesh(outletPoly, color="r")
    p.add_mesh(inletPoly, color="r")
    p.add_mesh(per_y_lower, color="r")
    p.add_mesh(per_y_upper, color="r")
    p.add_mesh(profilepoints2d, color="k")
    p.view_xy()
    p.screenshot(path)

    return sortedPoly, psPoly, ssPoly, per_y_upper, per_y_lower, inletPoly, outletPoly


def cascade_3d_domain(sortedPoly, psPoly, ssPoly, per_y_upper, per_y_lower, inletPoly, outletPoly, zspan, avdr=1,
                      path=None):
    if path is None:
        tmpdir = tempfile.mkdtemp()
        path = tmpdir + "/plot.png"

    x_lower = inletPoly.bounds[0]
    x_upper = outletPoly.bounds[0]

    def compute_transform(point, span, avdr, x_lower, x_upper, sign=1):
        l = abs(x_lower - x_upper)
        x = abs(point[0] - x_upper) / l
        return np.array([0, 0, sign * span * (1 + avdr * x)]) + point

    def transform(avdr, poly, x_lower, x_upper, zspan, sign):
        poly_copy = poly.copy()
        for idx, pt in enumerate(poly_copy.points):
            poly_copy.points[idx] = compute_transform(pt, zspan, avdr, x_lower, x_upper, sign)
        return poly_copy

    sortedPoly_lowz = transform(avdr, sortedPoly, x_lower, x_upper, zspan, -1)
    sortedPoly_high = transform(avdr, sortedPoly, x_lower, x_upper, zspan, 1)

    psPoly_lowz = transform(avdr, psPoly, x_lower, x_upper, zspan, -1)
    psPoly_highz = transform(avdr, psPoly, x_lower, x_upper, zspan, 1)

    ssPoly_lowz = transform(avdr, ssPoly, x_lower, x_upper, zspan, -1)
    ssPoly_highz = transform(avdr, ssPoly, x_lower, x_upper, zspan, 1)

    per_y_upper_lowz = transform(avdr, per_y_upper, x_lower, x_upper, zspan, -1)
    per_y_upper_highz = transform(avdr, per_y_upper, x_lower, x_upper, zspan, 1)

    per_y_lower_lowz = transform(avdr, per_y_lower, x_lower, x_upper, zspan, -1)
    per_y_lower_highz = transform(avdr, per_y_lower, x_lower, x_upper, zspan, 1)

    inletPoly_lowz = transform(avdr, inletPoly, x_lower, x_upper, zspan, -1)
    inletPoly_highz = transform(avdr, inletPoly, x_lower, x_upper, zspan, 1)

    outletPoly_lowz = transform(avdr, outletPoly, x_lower, x_upper, zspan, -1)
    outletPoly_highz = transform(avdr, outletPoly, x_lower, x_upper, zspan, 1)

    p = pv.Plotter(off_screen=True)
    p.add_mesh(psPoly_lowz, color="r")
    p.add_mesh(psPoly_highz, opacity=0.9)
    p.add_mesh(ssPoly_lowz, opacity=0.9)
    p.add_mesh(ssPoly_highz, opacity=0.9, color="white")
    p.add_mesh(per_y_upper_lowz, opacity=0.9, color="white")
    p.add_mesh(per_y_upper_highz, opacity=0.9, color="white")
    p.add_mesh(per_y_lower_lowz, opacity=0.9, color="white")
    p.add_mesh(per_y_lower_highz, opacity=0.9, color="white")
    p.add_mesh(inletPoly_lowz, opacity=0.9, color="white")
    p.add_mesh(inletPoly_highz, opacity=0.9, color="white")
    p.add_mesh(outletPoly_lowz, opacity=0.9, color="white")
    p.add_mesh(outletPoly_highz, opacity=0.9, color="white")
    p.view_xy()
    p.screenshot(path)

    return sortedPoly_lowz, sortedPoly_high, per_y_upper_lowz, per_y_upper_highz, \
           per_y_lower_lowz, per_y_lower_highz, \
           inletPoly_lowz, inletPoly_highz, \
           outletPoly_lowz, outletPoly_highz



import numpy as np

from ntrfc.fluid.turbulence import calcTu, calcTkeByTu, calcTke, calcFluc, calcRey


def test_calcTu():
    # Define sample inputs and expected output
    tke = 0.1
    Uabs = 2.0
    expected_output = np.sqrt(2.0 / 3.0 * tke) / max(1e-9, Uabs)

    # Call the function and compare with expected output
    actual_output = calcTu(tke, Uabs)
    assert np.isclose(actual_output, expected_output), f"Expected {expected_output}, but got {actual_output}"


def test_calcTkeByTu():
    # Test case 1
    Tu = 0.05
    Uabs = 1.0
    expected_tke = 0.0037500000000000007
    assert np.isclose(calcTkeByTu(Tu, Uabs), expected_tke)

    # Test case 2
    Tu = 0.0
    Uabs = 1.0
    expected_tke = 0.0
    assert np.isclose(calcTkeByTu(Tu, Uabs), expected_tke)


def test_calcTke():
    u_2 = 2.0
    v_2 = 3.0
    w_2 = 4.0
    expected_tke = 4.5  # calculated as 0.5 * (u_2 + v_2 + w_2)
    assert calcTke(u_2, v_2, w_2) == expected_tke


def test_calcFluc():
    series = np.random.randn(1000)  # calculated as 0.5 * (u_2 + v_2 + w_2)
    assert np.isclose(np.mean(calcFluc(series)), 0.0)
    series = np.random.randn(1000) + 1  # calculated as 0.5 * (u_2 + v_2 + w_2)
    assert np.isclose(np.mean(calcFluc(series)), 0.0)


def test_calcRey():
    series = np.stack([np.random.randn(20000), np.random.randn(20000), np.random.randn(20000)])

    uu, vv, ww, uv, uw, vw = calcRey(series[0], series[1], series[2])
    assert np.isclose(uu, 1.0, atol=5e-2)
    assert np.isclose(vv, 1.0, atol=5e-2)
    assert np.isclose(ww, 1.0, atol=5e-2)
    assert np.isclose(uv, 0.0, atol=5e-2)
    assert np.isclose(uw, 0.0, atol=5e-2)
    assert np.isclose(vw, 0.0, atol=5e-2)



import numpy as np

from ntrfc.timeseries.integral_scales import integralscales, get_self_correlating_frequencies


def test_integralscales():
    time = np.linspace(0, 100, 100000)
    freq = 1
    signal = np.sin(time * freq) + 1

    ts, ls = integralscales(signal, time)
    assert np.isclose(ts, 1, rtol=4e-3), "time scale not computed accurately"
    assert np.isclose(ls, 1, rtol=4e-3), "length scale not computed accurately"


def test_get_self_correlating_frequencies():
    # Test a sinusoidal signal with a frequency of 1Hz
    timesteps = np.linspace(0, 10, 1000)
    signal = np.sin(2 * np.pi * timesteps)
    frequencies = get_self_correlating_frequencies(timesteps, signal)
    assert np.isclose(frequencies[0], 1.0, rtol=1e-2)

    # Test a sinusoidal signal with a frequency of 2Hz
    timesteps = np.linspace(0, 5, 500)
    signal = np.sin(2 * np.pi * 2 * timesteps)
    frequencies = get_self_correlating_frequencies(timesteps, signal)
    assert np.isclose(frequencies[0], 2.0, rtol=1e-2)

    # Test a signal with multiple self-correlating frequencies
    timesteps = np.linspace(0, 20, 2000)
    signal = np.sin(2 * np.pi * timesteps) + np.sin(2 * np.pi * 2 * timesteps)
    frequencies = get_self_correlating_frequencies(timesteps, signal)
    assert np.isclose(frequencies[0], 2.0, rtol=1e-2)
    assert np.isclose(frequencies[1], 1.0, rtol=1e-2)



import numpy as np
from scipy import optimize
from scipy.spatial import Delaunay
from scipy.spatial import distance
from scipy.spatial.distance import cdist


def calc_concavehull(x, y, alpha):
    """
    origin: https://stackoverflow.com/questions/50549128/boundary-enclosing-a-given-set-of-points/50714300#50714300
    """
    points = []
    for i in range(len(x)):
        points.append([x[i], y[i]])

    points = np.asarray(points)

    def alpha_shape(points, alpha, only_outer=True):
        """
        Compute the alpha shape (concave hull) of a set of points.
        :param points: np.array of shape (n,2) points.
        :param alpha: alpha value.
        :param only_outer: boolean value to specify if we keep only the outer border
        or also inner edges.
        :return: set of (i,j) pairs representing edges of the alpha-shape. (i,j) are
        the indices in the points array.
        """

        assert points.shape[0] > 3, "Need at least four points"

        def add_edge(edges, i, j):
            """
            Add an edge between the i-th and j-th points,
            if not in the list already
            """
            if (i, j) in edges or (j, i) in edges:
                # already added
                assert (j, i) in edges, "Can't go twice over same directed edge right?"
                if only_outer:
                    # if both neighboring triangles are in shape, it's not a boundary edge
                    edges.remove((j, i))
                return
            edges.add((i, j))

        tri = Delaunay(points)
        edges = set()
        # Loop over triangles:
        # ia, ib, ic = indices of corner points of the triangle
        for ia, ib, ic in tri.vertices:
            pa = points[ia]
            pb = points[ib]
            pc = points[ic]
            # Computing radius of triangle circumcircle
            # www.mathalino.com/reviewer/derivation-of-formulas/derivation-of-formula-for-radius-of-circumcircle
            a = np.sqrt((pa[0] - pb[0]) ** 2 + (pa[1] - pb[1]) ** 2)
            b = np.sqrt((pb[0] - pc[0]) ** 2 + (pb[1] - pc[1]) ** 2)
            c = np.sqrt((pc[0] - pa[0]) ** 2 + (pc[1] - pa[1]) ** 2)
            s = (a + b + c) / 2.0

            A = (s * (s - a) * (s - b) * (s - c))
            if A > 0:
                area = np.sqrt(A)

                circum_r = a * b * c / (4.0 * area)
                if circum_r < alpha:
                    add_edge(edges, ia, ib)
                    add_edge(edges, ib, ic)
                    add_edge(edges, ic, ia)
        return edges

    def find_edges_with(i, edge_set):
        i_first = [j for (x, j) in edge_set if x == i]
        i_second = [j for (j, x) in edge_set if x == i]
        return i_first, i_second

    def stitch_boundaries(edges):
        edge_set = edges.copy()
        boundary_lst = []
        while len(edge_set) > 0:
            boundary = []
            edge0 = edge_set.pop()
            boundary.append(edge0)
            last_edge = edge0
            while len(edge_set) > 0:
                i, j = last_edge
                j_first, j_second = find_edges_with(j, edge_set)
                if j_first:
                    edge_set.remove((j, j_first[0]))
                    edge_with_j = (j, j_first[0])
                    boundary.append(edge_with_j)
                    last_edge = edge_with_j
                elif j_second:
                    edge_set.remove((j_second[0], j))
                    edge_with_j = (j, j_second[0])  # flip edge rep
                    boundary.append(edge_with_j)
                    last_edge = edge_with_j

                if edge0[0] == last_edge[1]:
                    break

            boundary_lst.append(boundary)
        return boundary_lst

    edges = alpha_shape(points, alpha)
    boundary_lst = stitch_boundaries(edges)
    x_new = []
    y_new = []

    if not len(boundary_lst) == 0:
        for i in range(len(boundary_lst[0])):
            x_new.append(points[boundary_lst[0][i][0]][0])
            y_new.append(points[boundary_lst[0][i][0]][1])

    return x_new, y_new


def calculate_minimal_distances(points):
    distances = cdist(points, points)  # Calculate pairwise distances
    np.fill_diagonal(distances, np.inf)  # Set diagonal elements to infinity
    minimal_distances = np.min(distances, axis=1)  # Find minimal distance for each point
    return minimal_distances


def auto_concaveHull(xs, ys):
    # Define the loss function
    def loss(alpha, xs, ys):
        xd, yd = calc_concavehull(xs, ys, alpha)
        if len(xd) == 0:
            return 1e10
        closed_points = np.stack([xd + [xd[0]], yd + [yd[0]]]).T
        points_orig = np.stack([xs, ys]).T
        centers = (closed_points[:-1] + closed_points[1:]) / 2
        distances = distance.cdist(centers, points_orig).min(axis=0)

        loss_distance_shape = np.max(distances)
        loss_norm_numpts = (len(xs) - len(xd)) / len(xs)

        return loss_distance_shape * 2 + loss_norm_numpts

    points = np.column_stack((xs, ys))
    # Compute the pairwise distances between all points
    distances = cdist(points, points)

    # Get the smallest and largest distances
    smallest_distance = np.min(distances[np.nonzero(distances)])  # Ignore zero distances
    largest_distance = np.max(distances)
    bounds_first_stage = [(smallest_distance, largest_distance / 2)]
    x0_first_stage = np.mean(bounds_first_stage, axis=1)

    result_first_stage = optimize.minimize(fun=loss, x0=x0_first_stage, args=(xs, ys,), method='Powell',
                                           bounds=bounds_first_stage).x[0]

    bounds_second_stage = [(result_first_stage / 2, result_first_stage)]

    result_second_stage = optimize.minimize(fun=loss, x0=result_first_stage * 0.9, args=(xs, ys,), method='Nelder-mead',
                                            bounds=bounds_second_stage).x[0]

    bounds_third_stage = [(1e-9, result_second_stage)]
    result = optimize.minimize(fun=loss, x0=result_second_stage, args=(xs, ys,), method='TNC',
                               bounds=bounds_third_stage).x[0]

    xans, yans = calc_concavehull(xs, ys, result)
    return xans, yans, result



def calc_inflow_cp(px, pt1, p1):
    """
    :param px: pressure at position
    :param pt1: total pressure inlet
    :param p1: pressure inlet
    :return: lift coefficient
    """
    cp = (px - pt1) / (p1 - pt1)
    return cp


# Totaldruckverlustbeiwert
def calc_zeta(pt1, pt2x, p2):
    """
    Calculates the Total Pressure Loss Coefficient (Zeta) for a fluid system.

    Parameters:
    -----------
    pt1 : float : Upstream total pressure [Pa]
    pt2x : float : Downstream total pressure [Pa]
    p2 : float : Downstream static pressure [Pa]

    Returns:
    --------
    zeta : float : Total Pressure Loss Coefficient (Zeta) [dimensionless]
    """
    zeta = (pt1 - pt2x) / (pt1 - p2)
    return zeta


def calc_cf(tau_w, u_inf, rho_inf):
    """
    Calculates skin friction coefficient

    :param tau_w: wall shear stress
    :type tau_w: float
    :param u_inf: freestream velocity
    :type u_inf: float
    :param rho_inf: freestream density
    :type rho_inf: float
    :return: skin friction coefficient
    :rtype: float
    """
    cf = 2 * tau_w / (rho_inf * u_inf ** 2)
    return cf



# Import modules:
from dataclasses import dataclass

import gmsh
import numpy as np
import pyvista as pv

from ntrfc.geometry.line import lines_from_points


# Initialize gmsh:


@dataclass
class MeshConfig:
    """
    Configuration for meshing
    """
    max_lc: float = 0.1  # Maximum characteristic length
    lc: float = 0.1  # Characteristic length for blade region
    bladeres: int = 300  # Number of elements in blade region
    bl_thickness: float = 0.1  # Thickness of boundary layer
    bl_growratio: float = 1.1  # Growth ratio of boundary layer
    bl_size: float = 0.1  # Size of boundary layer elements
    wake_length: float = 0.1  # Length of wake region
    wake_width: float = 0.1  # Width of wake region
    wake_lc: float = 0.1  # Characteristic length of wake region
    fake_yShiftCylinder: float = 0.1  # Shift cylinder to avoid wake region
    progression_le_halfss: float = 1.01
    progression_halfss_te: float = 0.99
    progression_te_halfps: float = 1.1
    progression_halfps_le: float = 0.9
    spansize: float = 0.1
    spanres: int = 5


def generate_turbocascade(domain2d,
                          meshconfig: MeshConfig,
                          filename: str,
                          verbose=False):
    """
    Generate a mesh for a turbocascade
    :param domain2d: Domain2D object
    :param meshconfig: MeshConfig object
    :param filename: Filename for mesh
    :param verbose: Print gmsh output
    :return:
    """

    # Initialize gmsh:
    gmsh.initialize()
    gmsh.model.add("cascade")
    points = {}
    splines = {}
    lines = {}
    curveloops = {}
    surfaceloops = {}
    surfaces = {}

    # Set meshing options:
    profilepoints_rolled_le = pv.PolyData(np.roll(domain2d.profilepoints.points, -domain2d.le_index, axis=0))
    le_index = 0
    te_index = (domain2d.te_index - domain2d.le_index) % domain2d.profilepoints.number_of_points

    domain2d.profilepoints_line = lines_from_points(profilepoints_rolled_le.points).compute_cell_sizes()

    lc_blade = np.sum(domain2d.profilepoints_line["Length"]) / meshconfig.bladeres

    sslengths = lines_from_points(domain2d.profilepoints_line.points[:te_index]).compute_cell_sizes()
    sslength = np.sum(sslengths["Length"])
    pslengths = lines_from_points(domain2d.profilepoints_line.points[te_index:]).compute_cell_sizes()
    pslength = np.sum(pslengths["Length"])

    ss_half_idx = np.where(np.cumsum(sslengths["Length"]) >= sslength / 2)[0][0]
    ps_half_idx = np.where(np.cumsum(pslengths["Length"]) >= pslength / 2)[0][0] + te_index

    # domain2d.profilepoints_spline = lines_from_points(domain2d.profilepoints.points)
    inlet_spline = lines_from_points(domain2d.inlet.points)
    domain2d.yperiodic_high_spline = lines_from_points(domain2d.yperiodic_high.points)
    outlet_spline = lines_from_points(domain2d.outlet.points[::-1])
    domain2d.yperiodic_low_spline = lines_from_points(domain2d.yperiodic_low.points[::-1])

    # domain2d.profilepoints_spline["ids"] = np.arange(domain2d.profilepoints_spline.number_of_points)
    inlet_spline["ids"] = np.arange(inlet_spline.number_of_points)
    domain2d.yperiodic_high_spline["ids"] = np.arange(domain2d.yperiodic_high_spline.number_of_points)
    outlet_spline["ids"] = np.arange(outlet_spline.number_of_points)
    domain2d.yperiodic_low_spline["ids"] = np.arange(domain2d.yperiodic_low_spline.number_of_points)

    # Create points and splines:

    points["blade"] = [gmsh.model.occ.add_point(*pt, lc_blade) for pt in profilepoints_rolled_le.points]
    points["domain2d.yperiodic_high"] = [gmsh.model.occ.add_point(*pt, meshconfig.lc) for pt in
                                         domain2d.yperiodic_high.points]
    points["domain2d.yperiodic_low"] = [gmsh.model.occ.add_point(*pt, meshconfig.lc) for pt in
                                        domain2d.yperiodic_low.points[::-1]]
    points["inlet"] = [points["domain2d.yperiodic_low"][-1], points["domain2d.yperiodic_high"][0]]
    points["outlet"] = [points["domain2d.yperiodic_high"][-1], points["domain2d.yperiodic_low"][0]]

    #    splines["blade"] = gmsh.model.occ.add_spline([*points["blade"], points["blade"][0]])
    splines["le_halfss"] = gmsh.model.occ.add_spline([*points["blade"][:ss_half_idx], points["blade"][ss_half_idx]])
    splines["halfss_te"] = gmsh.model.occ.add_spline(
        [*points["blade"][ss_half_idx:te_index], points["blade"][te_index]])
    splines["te_halfps"] = gmsh.model.occ.add_spline(
        [*points["blade"][te_index:ps_half_idx], points["blade"][ps_half_idx]])
    splines["halfps_le"] = gmsh.model.occ.add_spline([*points["blade"][ps_half_idx:], points["blade"][0]])

    splines["inlet"] = gmsh.model.occ.add_spline(points["inlet"])
    splines["domain2d.yperiodic_high"] = gmsh.model.occ.add_spline(points["domain2d.yperiodic_high"])
    splines["outlet"] = gmsh.model.occ.add_spline(points["outlet"])
    splines["domain2d.yperiodic_low"] = gmsh.model.occ.add_spline(points["domain2d.yperiodic_low"])

    curveloops["blade"] = gmsh.model.occ.add_curve_loop(
        [splines["le_halfss"], splines["halfss_te"], splines["te_halfps"], splines["halfps_le"]])
    curveloops["domain"] = gmsh.model.occ.add_curve_loop(
        [splines["inlet"], splines["domain2d.yperiodic_high"], splines["outlet"], splines["domain2d.yperiodic_low"]])

    surfaces["domain"] = gmsh.model.occ.add_plane_surface([curveloops["domain"], curveloops["blade"]])
    gmsh.model.occ.synchronize()

    # Boundary layer
    f = gmsh.model.mesh.field.add('BoundaryLayer')
    gmsh.model.mesh.field.setNumbers(f, 'CurvesList', [splines["le_halfss"], splines["halfss_te"], splines["te_halfps"],
                                                       splines["halfps_le"]])
    gmsh.model.mesh.field.setNumber(f, 'Size', meshconfig.bl_size)
    gmsh.model.mesh.field.setNumber(f, 'Ratio', meshconfig.bl_growratio)
    gmsh.model.mesh.field.setNumber(f, 'Quads', 1)
    gmsh.model.mesh.field.setNumber(f, 'Thickness', meshconfig.bl_thickness)
    gmsh.model.mesh.field.setAsBoundaryLayer(f)

    # blade resolution

    curvelength = sslength + pslength

    sscells = meshconfig.bladeres * sslength / curvelength
    pscells = meshconfig.bladeres * pslength / curvelength

    gmsh.model.mesh.set_transfinite_curve(splines["le_halfss"], int(sscells // 2), "Progression",
                                          meshconfig.progression_le_halfss)
    gmsh.model.mesh.set_transfinite_curve(splines["halfss_te"], int(sscells // 2), "Progression",
                                          meshconfig.progression_halfss_te)
    gmsh.model.mesh.set_transfinite_curve(splines["te_halfps"], int(pscells // 2), "Progression",
                                          meshconfig.progression_te_halfps)
    gmsh.model.mesh.set_transfinite_curve(splines["halfps_le"], int(pscells // 2), "Progression",
                                          meshconfig.progression_halfps_le)

    # Wake Resolution
    w = gmsh.model.mesh.field.add('Cylinder')
    gmsh.model.mesh.field.setNumber(w, "VIn", meshconfig.wake_lc)
    gmsh.model.mesh.field.setNumber(w, "VOut", meshconfig.max_lc)
    gmsh.model.mesh.field.setNumber(w, "Radius", meshconfig.wake_width)
    gmsh.model.mesh.field.setNumber(w, "XAxis", 0.5 * meshconfig.wake_length)

    minx = np.min(domain2d.profilepoints.points[::, 0])
    maxx = np.max(domain2d.profilepoints.points[::, 0])
    miny = np.min(domain2d.profilepoints.points[::, 1])
    maxy = np.max(domain2d.profilepoints.points[::, 1])

    wake_angle = np.deg2rad(domain2d.beta_trailing)
    gmsh.model.mesh.field.setNumber(w, "XCenter", minx + (maxx - minx) + 0.5 * meshconfig.wake_length)
    gmsh.model.mesh.field.setNumber(w, "YCenter",
                                    meshconfig.fake_yShiftCylinder + miny - 0.5 * meshconfig.wake_length * np.tan(
                                        wake_angle))
    gmsh.model.mesh.field.setNumber(w, "ZAxis", 0)
    gmsh.model.mesh.field.setNumber(w, "YAxis", -0.5 * meshconfig.wake_length * np.tan(wake_angle))
    gmsh.model.mesh.field.setNumber(w, "XAxis", 0.5 * meshconfig.wake_length)
    gmsh.model.mesh.field.setAsBackgroundMesh(w)
    gmsh.model.occ.synchronize()
    surfaceTags = gmsh.model.getEntities(2)


    gmsh.option.set_number("Mesh.Algorithm", 6)
    gmsh.option.setNumber("Mesh.ElementOrder", 1)
    # mesh options
    gmsh.option.setNumber("Mesh.RecombineAll", 1)
    gmsh.option.setNumber("Mesh.RecombinationAlgorithm", 1)
    # gmsh.option.setNumber("Mesh.SubdivisionAlgorithm", 2)  # all hexahedra

    gmsh.model.mesh.setOrder(1)
    gmsh.model.occ.synchronize()
    gmsh.model.occ.extrude(surfaceTags, dx=0, dy=0, dz=meshconfig.spansize,
                                     numElements=[meshconfig.spanres],
                                     recombine=True)
    gmsh.model.occ.synchronize()


    gmsh.model.addPhysicalGroup(3,[1], name="fluid")
    gmsh.model.addPhysicalGroup(2,[2], name="inlet")
    gmsh.model.addPhysicalGroup(2,[4], name="outlet")
    gmsh.model.addPhysicalGroup(2,[6,7,8,9], name="blade")
    gmsh.model.addPhysicalGroup(2,[1], name="z_lower")
    gmsh.model.addPhysicalGroup(2,[10], name="z_upper")
    gmsh.model.addPhysicalGroup(2,[3], name="y_upper")
    gmsh.model.addPhysicalGroup(2,[5], name="y_lower")

    gmsh.model.occ.synchronize()


    gmsh.model.mesh.generate(3)
    gmsh.model.occ.synchronize()

    gmsh.write(filename)
    gmsh.model.removePhysicalGroups(gmsh.model.occ.getEntities(0))
    gmsh.model.removePhysicalGroups(gmsh.model.occ.getEntities(1))
    gmsh.finalize()





def test_cascade_meshing_gmsh(tmpdir):
    import pyvista as pv
    import numpy as np
    from ntrfc.gmsh.turbo_cascade import generate_turbocascade, MeshConfig
    from ntrfc.turbo.airfoil_generators.naca_airfoil_creator import naca
    from ntrfc.cascade_case.utils.domain_utils import DomainParameters
    from ntrfc.cascade_case.domain import CascadeDomain2D
    from ntrfc.filehandling.mesh import load_mesh

    ptsx, ptsy = naca("6510", 200, False)
    # create a 3d pointcloud using pv.PolyData, all z values are 0
    pts = pv.PolyData(np.c_[ptsx, ptsy, np.zeros(len(ptsx))])
    domainparams = DomainParameters()
    domainparams.generate_params_by_pointcloud(pts)
    domainparams.xinlet = -1.4
    domainparams.xoutlet = 3
    domainparams.pitch = 1.6
    domainparams.blade_yshift = 0.05
    domain2d = CascadeDomain2D()
    domain2d.generate_from_cascade_parameters(domainparams)

    meshpath = tmpdir / "test.cgns"

    meshconfig = MeshConfig()

    di = 0.04

    meshconfig.max_lc = di
    meshconfig.min_lc = di / 10
    meshconfig.bl_thickness = di * 1.6
    meshconfig.bl_growratio = 1.2
    meshconfig.bl_size = 1.0e-5
    meshconfig.wake_length = domain2d.chordlength * 1
    meshconfig.wake_width = domain2d.chordlength * .1
    meshconfig.wake_lc = di * 0.5
    meshconfig.fake_yShiftCylinder = 0
    meshconfig.bladeres = int((domainparams.pspoly.length + domainparams.sspoly.length) / (meshconfig.min_lc * 4))
    meshconfig.progression_le_halfss = 1.05
    meshconfig.progression_halfss_te = 0.95
    meshconfig.progression_te_halfps = 1.05
    meshconfig.progression_halfps_le = 0.95
    meshconfig.spansize = 0.01
    meshconfig.spanres = 1

    generate_turbocascade(domain2d,
                          meshconfig,
                          str(meshpath), verbose=False)

    mesh = load_mesh(meshpath)

    assert mesh.number_of_cells > 0, "somethings wrong"



import numpy as np
import pyvista as pv


def test_createprofileprobes():
    from ntrfc.turbo.probegeneration import create_profileprobes
    from ntrfc.turbo.airfoil_generators.naca_airfoil_creator import naca
    from ntrfc.turbo.pointcloud_methods import extract_geo_paras
    naca_code = "6009"
    angle = 10  # deg
    res = 240
    xs, ys = naca(naca_code, res, half_cosine_spacing=False)
    sorted_poly = pv.PolyData(np.stack([xs, ys, np.zeros(len(xs))]).T)
    sorted_poly.rotate_z(angle, inplace=True)
    sorted_extracted_poly, psPoly, ssPoly, ind_vk, ind_hk, midsPoly, beta_leading, beta_trailing, camber_angle, alpha = extract_geo_paras(
        sorted_poly, 1)
    n_psprobes = 24
    n_ssprobes = 36
    probes_ss, probes_ps = create_profileprobes(ssPoly, psPoly, 1, n_ssprobes, n_psprobes, tolerance=1e-6)
    assert probes_ps.number_of_points == n_psprobes, "number of pressure side probes not correct"
    assert probes_ss.number_of_points == n_ssprobes, "number of suction side probes not correct"


def test_create_midpassageprobes():
    from ntrfc.turbo.probegeneration import create_midpassageprobes
    from ntrfc.turbo.airfoil_generators.naca_airfoil_creator import naca
    from ntrfc.turbo.pointcloud_methods import extract_geo_paras
    naca_code = "6009"
    angle = 10  # deg
    res = 240
    xs, ys = naca(naca_code, res, half_cosine_spacing=False)
    sorted_poly = pv.PolyData(np.stack([xs, ys, np.zeros(len(xs))]).T)
    sorted_poly.rotate_z(angle, inplace=True)
    sorted_extracted_poly, psPoly, ssPoly, ind_vk, ind_hk, midsPoly, beta_leading, beta_trailing, camber_angle, alpha = extract_geo_paras(
        sorted_poly, 1)
    nop = 40
    midspan_probes = create_midpassageprobes(1, -0.3, 0.3, 0.1, beta_leading, beta_trailing, midsPoly, nop)
    assert midspan_probes.number_of_points == nop, "number of probes on midpassage line not correct"


def test_stagnationpointprobes():
    from ntrfc.turbo.probegeneration import create_stagnationpointprobes
    from ntrfc.turbo.airfoil_generators.naca_airfoil_creator import naca

    naca_code = "6509"
    angle = 30  # deg
    res = 420
    xs, ys = naca(naca_code, res, half_cosine_spacing=False)
    sorted_poly = pv.PolyData(np.stack([xs[:-1], ys[:-1], np.zeros(len(xs) - 1)]).T)
    sorted_poly.rotate_z(angle, inplace=True)
    probes = create_stagnationpointprobes(1, 20, sorted_poly, 0, np.array([1, 0, 0]), 1)



#!/usr/bin/env python

"""The setup script."""

from pip._internal.req import parse_requirements
from setuptools import setup, find_packages

with open('README.rst') as readme_file:
    readme = readme_file.read()

with open('HISTORY.rst') as history_file:
    history = history_file.read()

install_reqs = parse_requirements("requirements.txt", session="hack")
requirements = [ir.requirement for ir in install_reqs]

test_requirements = ['pytest>=3', ]

setup(
    name='ntrfc',
    author="Malte Nyhuis",
    author_email='nyhuis@tfd.uni-hannover.de',
    python_requires='>=3.10',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3.10',
    ],
    description="Numerical Test Rig for Cascades. A workflows-library for cfd-analysis of cascade-flows",
    entry_points={
        'console_scripts': [
            'ntrfc=ntrfc.cli:main',
        ],
    },
    install_requires=requirements,
    license="MIT license",
    long_description=readme + '\n\n' + history,
    include_package_data=True,
    package_data={
        'ntrfc': ['data/*'],
    },
    keywords='ntrfc',
    packages=find_packages(include=['ntrfc', 'ntrfc.*']),
    test_suite='tests',
    tests_require=test_requirements,
    url='https://gitlab.uni-hannover.de/tfd_public/tools/NTRfC',
    version='0.1.7',
    zip_safe=False,
)



import numpy as np


def calcTu(tke, Uabs):
    # tke: turbulente kinetische Energie
    # Uabs: zeitlich gemittelte lokale Absolutgeschinwidgkeit

    Tu = np.sqrt(2.0 / 3.0 * tke) / max(1e-9, Uabs)

    return Tu


def calcTkeByTu(Tu, Uabs):
    Tke = (Tu * Uabs) ** 2 * (3 / 2.0)
    return Tke


def calcTke(u_2, v_2, w_2):
    # u_2: u-fluktuationen quadriert und gemittelt
    # v_2: v-fluktuationen quadriert und gemittelt
    # w_2: w-fluktuationen quadriert und gemittelt
    tke = (0.5 * (u_2 + v_2 + w_2))
    return tke


def calcFluc(velo):
    mean_velo = np.mean(velo)

    velo = np.array(velo)

    return velo - mean_velo


def calcRey(u, v, w):
    u_ = calcFluc(u)
    v_ = calcFluc(v)
    w_ = calcFluc(w)

    u_u_ = np.multiply(u_, u_)
    v_v_ = np.multiply(v_, v_)
    w_w_ = np.multiply(w_, w_)
    u_v_ = np.multiply(u_, v_)
    u_w_ = np.multiply(u_, w_)
    v_w_ = np.multiply(v_, w_)

    mean_u_u_ = np.mean(u_u_)
    mean_v_v_ = np.mean(v_v_)
    mean_w_w_ = np.mean(w_w_)
    mean_u_v_ = np.mean(u_v_)
    mean_u_w_ = np.mean(u_w_)
    mean_v_w_ = np.mean(v_w_)

    return [mean_u_u_, mean_v_v_, mean_w_w_, mean_u_v_, mean_u_w_, mean_v_w_]



import numpy as np


# import pyvista as pv


def calc_dist_from_surface(surface_primary, surface_secondary, verbose=False):
    """
    Distance Between Two Surfaces / A Surface and a Pointcloud
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    Compute the average thickness between two surfaces.

    We can compute the thickness between the two surfaces using a few different
    methods. We will demo a method where we compute the normals of the
    bottom surface, and then project a ray to the top surface to compute the
    distance along the surface normals.
    :param surface_primary: pv.UnstructuredGrid
    :param surface_secondary: pv.UnstructuredGrid / pv.PolyData
    :param verbose: plots?
    :return: surface_primary with distances from secondary
    """

    ###############################################################################
    # Ray Tracing Distance
    # ++++++++++++++++++++
    #

    h0n = surface_primary.compute_normals(point_normals=True, cell_normals=False,
                                          auto_orient_normals=True)

    ###############################################################################
    # Travel along normals to the other surface and compute the thickness on each
    # vector.

    h0n["distances"] = np.empty(surface_primary.n_points)
    for i in range(h0n.n_points):
        p = h0n.points[i]
        vec = h0n["Normals"][i] * h0n.length
        p0 = p - vec
        p1 = p + vec
        ip, ic = surface_secondary.ray_trace(p0, p1, first_point=True)
        dist = np.sqrt(np.sum((ip - p) ** 2))
        h0n["distances"][i] = dist

    if any(h0n["distances"] == 0):
        # Replace zeros with nans
        mask = h0n["distances"] == 0
        h0n["distances"][mask] = np.nan
        np.nanmean(h0n["distances"])

    # if verbose:
    #     ###############################################################################
    #     p = pv.Plotter()
    #     p.add_mesh(h0n, scalars="distances", smooth_shading=True)
    #     p.add_mesh(surface_secondary, color=True, opacity=0.75, smooth_shading=True)
    #     p.show()

    return h0n



import numpy as np


def compute_cell_skewness(grid):
    """
    Compute the skewness of each cell in an unstructured grid.

    Parameters
    ----------
    grid : pyvista.UnstructuredGrid
        The input unstructured grid.

    Returns
    -------
    skewness : list
        A list of the skewness of each cell in the grid.
    """
    # Get the centroids of each cell

    # Compute the skewness of each cell
    skewnesses = []
    for i in range(grid.n_cells):
        cell = grid.extract_cells(i)
        skewnesses.append(skewness(cell.points))

    return skewnesses


def skewness(points):
    """
    Compute the skewness of a set of points.

    The skewness is defined as the difference between the maximum and minimum
    distances from the centroid of the points to the vertices of the cell,
    divided by the median distance from the centroid to the vertices of the
    cell.

    Parameters
    ----------
    points : numpy array
        The points of the cell, represented as a 2D array with shape (n, 3),
        where n is the number of points and the columns are the x, y, and z
        coordinates.

    Returns
    -------
    skewness : float
        The skewness of the points.
    """
    # Compute the centroid of the points
    centroid = np.mean(points, axis=0)

    # Compute the distances from the centroid to the vertices of the cell
    distances = np.linalg.norm(points - centroid, axis=1)

    # Compute the skewness as the difference between the maximum and minimum
    # distances, divided by the median distance
    skewness = (np.max(distances) - np.min(distances)) / np.median(distances)

    return skewness



def test_getgci():
    from ntrfc.meshquality.grid_convergece_index import getGCI
    import numpy as np
    #
    fc3 = 23.263
    fc2 = 23.165
    fc1 = 23.151

    N3 = 18576
    N2 = 74304
    N1 = 297216

    Fs = 1.25
    D = 2

    GCI_1, GCI_2, GCI_3 = getGCI(N1, N2, N3, fc1, fc2, fc3, D, Fs=1.25)

    assert np.isclose(GCI_1, 0.0001259844787122061), "GCI_1 computation does not seem to be correct"
    assert np.isclose(GCI_2, 0.0008813583711057829), "GCI_3 computation does not seem to be correct"
    assert np.isclose(GCI_3, 0.0061695085977409286), "GCI_3 computation does not seem to be correct"



import numpy as np

from ntrfc.fluid.fluid import mach_number, total_pressure


def p_t_is_from_mach(kappa, mach_number, static_pressure):
    # Calculates total pressure in isentropic flow
    # Source: https://www.grc.nasa.gov/www/BGH/isentrop.html
    total_pressure = static_pressure * pow(1.0 + (kappa - 1.0) / 2.0 * pow(mach_number, 2.0), (kappa / (kappa - 1.0)))
    return total_pressure


def p_is_from_mach(kappa, ma, p_t_is):
    # Calculates static pressure in isentropic flow
    # Source: https://www.grc.nasa.gov/www/BGH/isentrop.html
    p_is = p_t_is / pow(1.0 + (kappa - 1.0) / 2.0 * pow(ma, 2.0), (kappa / (kappa - 1.0)))
    return p_is


def temp_t_is(kappa, ma, T):
    # Calculates total temperature in isentropic flow
    # Source: https://www.grc.nasa.gov/www/BGH/isentrop.html
    T_t_is = T / (((1.0 + (kappa - 1.0) * 0.5 * ma ** 2.0)) ** (-1.0))
    return T_t_is


def temp_is(kappa, ma, Tt):
    # Calculates static temperature in isentropic flow
    # Source: https://www.grc.nasa.gov/www/BGH/isentrop.html
    T = Tt / (1 + ((kappa - 1) / 2.0) * ma ** 2)
    return T


def mach_is_x(kappa, p, pt):
    # Calculates local isentropic Mach number
    if p > 0 and pt > 0 and pt >= p:
        y = np.sqrt(2 / (kappa - 1) * ((pt / p) ** ((kappa - 1) / kappa) - 1))
    else:
        y = 0.0
    return y


def isentropic_mach_number(isentropic_pressure, kappa, static_pressure, mach, gas_constant, static_temperature):
    """
    Calculates the isentropic Mach number.

    Parameters:
        isentropic_pressure (float): Isentropic pressure of the flow.
        kappa (float): Specific heat ratio of the gas.
        static_pressure (float): Static pressure of the flow.
        mach (float): Mach number of the flow.
        gas_constant (float): Gas constant of the gas.
        static_temperature (float): Static temperature of the flow.

    Returns:
        float: Isentropic Mach number.

    """
    # Calculate the total pressure
    total_pressure = p_t_is_from_mach(kappa, mach_number(mach, kappa, gas_constant, static_temperature),
                                      static_pressure)

    # Calculate the dynamic pressure
    dynamic_pressure = total_pressure - isentropic_pressure

    # Calculate the isentropic Mach number
    isentropic_mach_number = np.sqrt(
        2.0 / (kappa - 1.0) * (pow(1.0 + (dynamic_pressure / isentropic_pressure), (kappa - 1.0) / kappa) - 1.0))

    return isentropic_mach_number


def ma_is(outflow_static_pressure, isentropic_exponent, pressure, velocity, gas_constant, temperature):
    """
    Calculate the isentropic Mach number at the outflow of a system.

    Parameters:
        outflow_static_pressure (float): static pressure at the outflow (Pa)
        isentropic_exponent (float): isentropic exponent of the gas
        pressure (float): pressure at a reference point in the flow (Pa)
        density (float): density at the reference point (kg/m^3)
        velocity (float): velocity at the reference point (m/s)
        gas_constant (float): gas constant of the gas (J/kg*K)
        temperature (float): temperature at the reference point (K)

    Returns:
        float: isentropic Mach number at the outflow
    """
    reference_point_total_pressure = total_pressure(isentropic_exponent,
                                                    mach_number(velocity, isentropic_exponent, gas_constant,
                                                                temperature), pressure)
    q2th = reference_point_total_pressure - outflow_static_pressure
    isentropic_mach_number = np.sqrt(2.0 / (isentropic_exponent - 1.0) * (
        pow(1.0 + (q2th / outflow_static_pressure), (isentropic_exponent - 1.0) / isentropic_exponent) - 1.0))
    return isentropic_mach_number


def isentropic_reynolds_number(kappa, specific_gas_constant, chord_length, sutherland_reference_viscosity,
                               mach_number, pressure, temperature,
                               sutherland_reference_temperature):
    """
    Calculates the isentropic Reynolds number at a point in a gas flow.

    Parameters:
    - kappa: the ratio of specific heats for the gas
    - specific_gas_constant: the specific gas constant for the gas
    - chord_length: the chord length of the body or structure
    - sutherland_reference_viscosity: the Sutherland reference viscosity
    - mach_number: the Mach number at the point
    - pressure: the pressure at the reference point
    - temperature: the temperature at the reference point
    - velocity_magnitude: the velocity magnitude at the reference point
    - isobaric_heat_constant: the isobaric heat constant
    - sutherland_reference_temperature: the Sutherland reference temperature

    Returns:
    - the isentropic Reynolds number at the point
    """
    total_temperature = isentropic_total_temperature(kappa, mach_number, temperature)
    iso_temperature = total_temperature / (1 + (kappa - 1) / 2 * mach_number ** 2)
    y = (kappa / specific_gas_constant) ** 0.5 * chord_length / sutherland_reference_viscosity * (
        mach_number * pressure * (iso_temperature + sutherland_reference_temperature) / iso_temperature ** 2)
    return y


def isentropic_total_temperature(kappa, mach_number, temperature):
    """
    Calculates the isentropic total temperature at a point in a gas.

    https://www.grc.nasa.gov/www/BGH/isentrop.html
    Eq #7

    Parameters:
    - kappa: the ratio of specific heats for the gas
    - mach_number: the Mach number at the point
    - temperature: the temperature at the point

    Returns:
    - the isentropic total temperature at the point
    """
    isentropic_total_temperature = temperature / (1 + (kappa - 1) / 2 * mach_number ** 2) ** -1
    return isentropic_total_temperature



import numpy as np
import pyvista as pv

from ntrfc.meshquality.nondimensionals import compute_scalar_gradient, constructWallMesh, cellDirections


def test_cellspans():
    from ntrfc.meshquality.nondimensionals import cellSpans
    # 11**3 nodes are enough
    height = 2
    width = 1
    length = 1
    nodes = 11
    # needs to be something simple
    mu_0 = 1  # dynamic viscosity
    rho = 1
    velocity = 2
    gradient = velocity / height
    # analytical solution_utils
    span_x = width / (nodes - 1)
    span_y = height / (nodes - 1)
    span_z = length / (nodes - 1)
    # define the mesh
    grid = structured_testgrid(height, length, nodes, width)

    # define velocityfield
    bounds = grid.bounds
    min_z = bounds[4]
    grid["U"] = [gradient * (grid.cell_centers().points[::, 1][i] - min_z) * np.array([0, 0, 1]) for i in
                 range(grid.number_of_cells)]
    grid["rho"] = np.ones(grid.number_of_cells)
    # extract surface
    surface = grid.extract_surface()
    facecellids = surface.surface_indices()
    upperwallids = []
    lowerwallids = []
    for faceid in facecellids:
        cell = surface.extract_cells(faceid)
        if all(cell.points[::, 1] == 0):
            lowerwallids.append(faceid)
        elif all(cell.points[::, 1] == height):
            upperwallids.append(faceid)
    lowerwall = surface.extract_cells(lowerwallids)
    upperwall = surface.extract_cells(upperwallids)
    lowerwall["U"] = [gradient * (lowerwall.cell_centers().points[::, 1][i] - min_z) * np.array([0, 0, 1]) for i in
                      range(lowerwall.number_of_cells)]
    upperwall["U"] = [gradient * (upperwall.cell_centers().points[::, 1][i] - min_z) * np.array([0, 0, 1]) for i in
                      range(upperwall.number_of_cells)]
    lowerwall["rho"] = np.ones(lowerwall.number_of_cells) * rho
    upperwall["rho"] = np.ones(upperwall.number_of_cells) * rho
    surface = constructWallMesh([lowerwall, upperwall])
    surfacenormals_surface = surface.extract_surface().compute_normals()
    walladjacentids = grid.find_containing_cell(surfacenormals_surface.points)
    volmesh_walladjacent = grid.extract_cells(walladjacentids)

    volmesh_walladjacent["cellCenters"] = volmesh_walladjacent.cell_centers().points
    volmesh_walladjacent["wallNormal"] = [
        surfacenormals_surface.point_data["Normals"][surfacenormals_surface.find_closest_point(i)]
        for i in volmesh_walladjacent.points]
    spans = np.array(cellSpans(volmesh_walladjacent, "U"))

    assert np.all(np.isclose(spans[::, 0], span_x))
    assert np.all(np.isclose(spans[::, 1], span_y))
    assert np.all(np.isclose(spans[::, 2], span_z))


def test_calc_dimensionless_gridspacing():
    """
    this method tests the nondimensional gridspacing postprocessing function calc_dimensionless_gridspacing
    a volume mesh will be created. boundary meshes will be extracted of the volume mesh.
    then a simple couette-velocity-field is defined

    calc_dimensionless_gridspacing needs to compute accurately the Delta+-Values

    """
    from ntrfc.meshquality.nondimensionals import calc_dimensionless_gridspacing
    import numpy as np

    def runtest(height, length, width):
        # 11**3 nodes are enough
        nodes = 11
        # needs to be something simple
        mu_0 = 1  # dynamic viscosity
        rho = 1
        velocity = 2
        gradient = velocity / height
        # analytical solution_utils
        span_x = width / (nodes - 1)
        span_y = height / (nodes - 1)
        span_z = length / (nodes - 1)
        wallshearstress = mu_0 * gradient
        wallshearvelocity = np.sqrt(wallshearstress / rho)
        deltaxplus = wallshearvelocity * span_x / mu_0
        deltayplus = wallshearvelocity * span_y / mu_0
        deltazplus = wallshearvelocity * span_z / mu_0
        grid = structured_testgrid(height, length, nodes, width)

        # define velocityfield
        bounds = grid.bounds
        min_z = bounds[4]
        grid["U"] = [gradient * (grid.cell_centers().points[::, 1][i] - min_z) * np.array([0, 0, 1]) for i in
                     range(grid.number_of_cells)]
        grid["rho"] = np.ones(grid.number_of_cells)
        # extract surface
        surface = grid.extract_surface()
        facecellids = surface.surface_indices()
        upperwallids = []
        lowerwallids = []
        for faceid in facecellids:
            cell = surface.extract_cells(faceid)
            if all(cell.points[::, 1] == 0):
                lowerwallids.append(faceid)
            elif all(cell.points[::, 1] == height):
                upperwallids.append(faceid)
        lowerwall = surface.extract_cells(lowerwallids)
        upperwall = surface.extract_cells(upperwallids)
        lowerwall["U"] = [gradient * (lowerwall.cell_centers().points[::, 1][i] - min_z) * np.array([0, 0, 1]) for i in
                          range(lowerwall.number_of_cells)]
        upperwall["U"] = [gradient * (upperwall.cell_centers().points[::, 1][i] - min_z) * np.array([0, 0, 1]) for i in
                          range(upperwall.number_of_cells)]
        lowerwall["rho"] = np.ones(lowerwall.number_of_cells) * rho
        upperwall["rho"] = np.ones(upperwall.number_of_cells) * rho
        dimless_gridspacings = calc_dimensionless_gridspacing(grid, [lowerwall, upperwall], "U", "rho", mu_0)
        assert all(np.isclose(deltazplus, dimless_gridspacings[
            "DeltaXPlus"], rtol=0.05)), "calc_dimensionelss_gridspcing in x direction not accurate"
        assert all(np.isclose(deltayplus, dimless_gridspacings[
            "DeltaYPlus"], rtol=0.05)), "calc_dimensionelss_gridspcing in y direction not accurate"
        assert all(np.isclose(deltaxplus, dimless_gridspacings[
            "DeltaZPlus"], rtol=0.05)), "calc_dimensionelss_gridspcing in z direction not accurate"

    runtest(height=1, length=1, width=1)
    runtest(height=2, length=1, width=1)
    runtest(height=1, length=2, width=1)
    runtest(height=1, length=1, width=2)


def structured_testgrid(height, length, nodes, width):
    # define the mesh
    xrng = np.arange(0, nodes, 1, dtype=np.float32)
    yrng = np.arange(0, nodes, 1, dtype=np.float32)
    zrng = np.arange(0, nodes, 1, dtype=np.float32)
    x, y, z = np.meshgrid(xrng, yrng, zrng)
    grid = pv.StructuredGrid(x, y, z)
    # scale the mesh
    grid.points /= nodes - 1
    grid.points *= np.array([width, height, length])
    return grid


def test_calc_dimensionless_gridspacing():
    """
    this method tests the nondimensional gridspacing postprocessing function calc_dimensionless_gridspacing
    a volume mesh will be created. boundary meshes will be extracted of the volume mesh.
    then a simple couette-velocity-field is defined

    calc_dimensionless_gridspacing needs to compute accurately the Delta+-Values

    """
    from ntrfc.meshquality.nondimensionals import calc_dimensionless_yplus
    import numpy as np

    def runtest(height, length, width):
        # 11**3 nodes are enough
        nodes = 11
        # needs to be something simple
        mu_0 = 1  # dynamic viscosity
        rho = 1
        velocity = 2
        gradient = velocity / height
        # analytical solution_utils
        span_x = width / (nodes - 1)
        span_y = height / (nodes - 1)
        span_z = length / (nodes - 1)
        wallshearstress = mu_0 * gradient
        wallshearvelocity = np.sqrt(wallshearstress / rho)
        deltayplus = wallshearvelocity * span_y / 2 / mu_0
        # define the mesh
        grid = structured_testgrid(height, length, nodes, width)

        # define velocityfield
        bounds = grid.bounds
        min_z = bounds[4]
        grid["U"] = [gradient * (grid.cell_centers().points[::, 1][i] - min_z) * np.array([0, 0, 1]) for i in
                     range(grid.number_of_cells)]
        grid["rho"] = np.ones(grid.number_of_cells)
        # extract surface
        surface = grid.extract_surface()
        facecellids = surface.surface_indices()
        upperwallids = []
        lowerwallids = []
        for faceid in facecellids:
            cell = surface.extract_cells(faceid)
            if all(cell.points[::, 1] == 0):
                lowerwallids.append(faceid)
            elif all(cell.points[::, 1] == height):
                upperwallids.append(faceid)
        lowerwall = surface.extract_cells(lowerwallids)
        upperwall = surface.extract_cells(upperwallids)
        lowerwall["U"] = [gradient * (lowerwall.cell_centers().points[::, 1][i] - min_z) * np.array([0, 0, 1]) for i in
                          range(lowerwall.number_of_cells)]
        upperwall["U"] = [gradient * (upperwall.cell_centers().points[::, 1][i] - min_z) * np.array([0, 0, 1]) for i in
                          range(upperwall.number_of_cells)]
        lowerwall["rho"] = np.ones(lowerwall.number_of_cells) * rho
        upperwall["rho"] = np.ones(upperwall.number_of_cells) * rho
        dimless_gridspacings = calc_dimensionless_yplus(grid, [lowerwall, upperwall], "U", "rho", mu_0)
        assert all(np.isclose(deltayplus, dimless_gridspacings[
            "yPlus"], rtol=0.05)), "calc_dimensionelss_gridspcing in x direction not accurate"

    runtest(height=1, length=1, width=1)
    runtest(height=2, length=1, width=1)
    runtest(height=1, length=2, width=1)
    runtest(height=1, length=1, width=2)


def test_compute_wallshearstress():
    from ntrfc.meshquality.nondimensionals import get_wall_shear_stress_velocity
    # 11**3 nodes are enough
    nodes = 11
    # needs to be something simple
    mu_0 = 1  # dynamic viscosity
    rho = 1
    velocity = 2
    height = 2
    length = 1
    width = 1
    gradient = velocity / height
    # analytical solution_utils
    wallshearstress = mu_0 * gradient
    wallshearvelocity = np.sqrt(wallshearstress / rho)

    # define the mesh
    grid = structured_testgrid(height, length, nodes, width)

    # define velocityfield
    bounds = grid.bounds
    min_z = bounds[4]
    grid["U"] = [gradient * (grid.cell_centers().points[::, 1][i] - min_z) * np.array([0, 0, 1]) for i in
                 range(grid.number_of_cells)]
    grid["rho"] = np.ones(grid.number_of_cells)
    grid = compute_scalar_gradient(grid, "U")

    # extract surface
    surface = grid.extract_surface()
    facecellids = surface.surface_indices()
    upperwallids = []
    lowerwallids = []
    for faceid in facecellids:
        cell = surface.extract_cells(faceid)
        if all(cell.points[::, 1] == 0):
            lowerwallids.append(faceid)
        elif all(cell.points[::, 1] == height):
            upperwallids.append(faceid)
    lowerwall = surface.extract_cells(lowerwallids)
    upperwall = surface.extract_cells(upperwallids)

    lowerwall["U"] = [gradient * (lowerwall.cell_centers().points[::, 1][i] - min_z) * np.array([0, 0, 1]) for i in
                      range(lowerwall.number_of_cells)]
    upperwall["U"] = [gradient * (upperwall.cell_centers().points[::, 1][i] - min_z) * np.array([0, 0, 1]) for i in
                      range(upperwall.number_of_cells)]
    lowerwall["rho"] = np.ones(lowerwall.number_of_cells) * rho
    upperwall["rho"] = np.ones(upperwall.number_of_cells) * rho

    surface = constructWallMesh([lowerwall, upperwall])
    surfacenormals_surface = surface.extract_surface().compute_normals()

    walladjacentids = grid.find_containing_cell(surfacenormals_surface.points)
    volmesh_walladjacent = grid.extract_cells(walladjacentids)

    volmesh_walladjacent["cellCenters"] = volmesh_walladjacent.cell_centers().points
    volmesh_walladjacent["wallNormal"] = [
        surfacenormals_surface.point_data["Normals"][surfacenormals_surface.find_closest_point(i)]
        for i in volmesh_walladjacent.points]

    utaus = get_wall_shear_stress_velocity(volmesh_walladjacent, mu_0, "rho", "U")

    assert np.all(np.isclose(utaus, wallshearvelocity))


def test_compute_scalar_gradient():
    # 11**3 nodes are enough
    nodes = 11
    # needs to be something simple
    height = 1
    length = 1
    width = 1
    velocity = 2
    gradient = velocity / height

    grid = structured_testgrid(height, length, nodes, width)
    # define velocityfield
    bounds = grid.bounds
    min_z = bounds[4]
    grid["U"] = [gradient * (grid.cell_centers().points[::, 1][i] - min_z) * np.array([0, 0, 1]) for i in
                 range(grid.number_of_cells)]
    expected = np.array([[0., 0., 0.],
                         [0., 0., 0.],
                         [0., 2., 0.]])
    # Compute the gradient and test it
    mesh = compute_scalar_gradient(grid, "U")
    grad_array = mesh[f"grad_U"].reshape(mesh.number_of_cells, 3, 3)
    assert np.all([np.all(np.isclose(expected, i)) for i in grad_array])


def test_constructWallMesh():
    # Create some test surfaces to merge
    surface_1 = pv.Plane()
    surface_2 = pv.Plane((1, 1, 1))

    # Call the function with the test surfaces
    result = constructWallMesh([surface_1, surface_2])

    # Check that the result is a valid UnstructuredGrid object
    assert isinstance(result, pv.UnstructuredGrid)

    # Check that the result has the expected number of points and cells
    assert result.n_points == 242
    assert result.n_cells == 200


def test_cellDirections():
    # Test 1: Check that unit vectors are returned correctly
    cellUMean = np.array([1, 2, 3])
    wallNorm = np.array([4, 5, 6])
    expected = np.array([[0.26726124, 0.53452248, 0.80178373], [0.45584231, 0.56980288, 0.68376346],
                         [-0.40824829, 0.81649658, -0.40824829]])
    actual = cellDirections(cellUMean, wallNorm)
    assert np.all(np.isclose(actual, expected))



def sutherland_viscosity(temperature, sutherland_constant=1.458e-06, reference_temperature=110.4):
    """
    Calculate the dynamic viscosity of a gas using Sutherland's Law.

    Parameters:
    - temperature (float): the temperature of the gas in degrees Kelvin
    - sutherland_constant (float): the Sutherland's constant for the gas (default value is for air)
    - reference_temperature (float): the reference temperature for the gas, in degrees Kelvin (default value is for air)

    Returns:
    - dynamic_viscosity (float): the dynamic viscosity of the gas, in units of square meters per second
    """

    dynamic_viscosity = sutherland_constant * (temperature) ** (0.5) / (1 + reference_temperature / temperature)
    return dynamic_viscosity


def total_pressure(kappa, mach_number, pressure):
    """
    Calculates the total pressure at a point in a gas using the isentropic flow equations.
    https://www.grc.nasa.gov/www/BGH/isentrop.html

    Parameters:
    - kappa: the ratio of specific heats for the gas at the point
    - mach_number: the Mach number at the point
    - pressure: the pressure at the point

    Returns:
    - the total pressure at the point
    """
    total_pressure = pressure * (1 + (kappa - 1) / 2 * mach_number ** 2) ** (kappa / (kappa - 1))
    return total_pressure


def mach_number(speed, specific_heat_ratio, specific_gas_constant, temperature):
    """
    Calculates the Mach number at a point in a gas.

    Parameters:
    - speed: the speed of the gas
    - specific_heat_ratio: the ratio of specific heats for the gas
    - specific_gas_constant: the specific gas constant for the gas
    - temperature: the temperature of the gas

    Returns:
    - the Mach number at the point
    """
    mach_number = speed / ((specific_heat_ratio * specific_gas_constant * temperature) ** 0.5)
    return mach_number



# Generate and plot the contour of an airfoil
# using the PARSEC parameterization

# Repository & documentation:
# http://github.com/dqsis/parsec-airfoils
# -------------------------------------


# Import libraries
from __future__ import division

from math import sqrt, tan, pi

import numpy as np


# User function pcoef
def pcoef(xte, yte, rle, x_cre, y_cre, d2ydx2_cre, th_cre, surface):
    """
    evaluate the PARSEC coefficients
    """

    # Initialize coefficients
    coef = np.zeros(6)

    # 1st coefficient depends on surface (pressure or suction)
    if surface.startswith('p'):
        coef[0] = -sqrt(2 * rle)
    else:
        coef[0] = sqrt(2 * rle)

    # Form system of equations
    A = np.array([
        [xte ** 1.5, xte ** 2.5, xte ** 3.5, xte ** 4.5, xte ** 5.5],
        [x_cre ** 1.5, x_cre ** 2.5, x_cre ** 3.5, x_cre ** 4.5,
         x_cre ** 5.5],
        [1.5 * sqrt(xte), 2.5 * xte ** 1.5, 3.5 * xte ** 2.5,
         4.5 * xte ** 3.5, 5.5 * xte ** 4.5],
        [1.5 * sqrt(x_cre), 2.5 * x_cre ** 1.5, 3.5 * x_cre ** 2.5,
         4.5 * x_cre ** 3.5, 5.5 * x_cre ** 4.5],
        [0.75 * (1 / sqrt(x_cre)), 3.75 * sqrt(x_cre), 8.75 * x_cre ** 1.5,
         15.75 * x_cre ** 2.5, 24.75 * x_cre ** 3.5]
    ])

    B = np.array([
        [yte - coef[0] * sqrt(xte)],
        [y_cre - coef[0] * sqrt(x_cre)],
        [tan(th_cre * pi / 180) - 0.5 * coef[0] * (1 / sqrt(xte))],
        [-0.5 * coef[0] * (1 / sqrt(x_cre))],
        [d2ydx2_cre + 0.25 * coef[0] * x_cre ** (-1.5)]
    ])

    # Solve system of linear equations
    X = np.linalg.solve(A, B)

    # Gather all coefficients
    coef[1:6] = X[0:5, 0]

    # Return coefficients
    return coef


def parsec_airfoil_gen(pparray, halfsinespacing=True, resolution=2000):
    # TE & LE of airfoil (normalized, chord = 1)
    xle = 0.0
    # yle = 0.0
    xte = 1.0
    yte = 0.0

    # LE radius
    rle = pparray[0]

    # Pressure (lower) surface parameters
    x_pre = pparray[1]
    y_pre = pparray[2]
    d2ydx2_pre = pparray[3]
    th_pre = pparray[4]

    # Suction (upper) surface parameters
    x_suc = pparray[5]
    y_suc = pparray[6]
    d2ydx2_suc = pparray[7]
    th_suc = pparray[8]

    # Evaluate pressure (lower) surface coefficients
    cf_pre = pcoef(xte, yte, rle,
                   x_pre, y_pre, d2ydx2_pre, th_pre,
                   'pre')

    # Evaluate suction (upper) surface coefficients
    cf_suc = pcoef(xte, yte, rle,
                   x_suc, y_suc, d2ydx2_suc, th_suc,
                   'suc')

    # Evaluate pressure (lower) surface points

    if halfsinespacing:
        beta = np.linspace(0.0, pi, resolution)
        halfsinespacing = [(0.5 * (1.0 - np.cos(x))) for x in beta]
        xx_pre = np.array(halfsinespacing[::-1])
        # xx_suc = np.array(halfsinespacing)

    else:  # Evaluate pressure (lower) surface points
        xx_pre = np.linspace(xte, xle, 101)
        # Evaluate suction (upper) surface points
        # xx_suc = np.linspace(xle, xte, 101)

    yy_pre = (cf_pre[0] * xx_pre ** (1 / 2) +
              cf_pre[1] * xx_pre ** (3 / 2) +
              cf_pre[2] * xx_pre ** (5 / 2) +
              cf_pre[3] * xx_pre ** (7 / 2) +
              cf_pre[4] * xx_pre ** (9 / 2) +
              cf_pre[5] * xx_pre ** (11 / 2)
              )

    # Evaluate suction (upper) surface points
    # xx_suc = x * np.linspace(xle, xte, res)
    xx_suc = np.array(halfsinespacing)
    yy_suc = (cf_suc[0] * xx_suc ** (1 / 2) +
              cf_suc[1] * xx_suc ** (3 / 2) +
              cf_suc[2] * xx_suc ** (5 / 2) +
              cf_suc[3] * xx_suc ** (7 / 2) +
              cf_suc[4] * xx_suc ** (9 / 2) +
              cf_suc[5] * xx_suc ** (11 / 2)
              )

    ps_points = np.stack([xx_pre, yy_pre, np.zeros(len(xx_pre))]).T
    ss_points = np.stack([xx_suc, yy_suc, np.zeros(len(xx_suc))]).T
    return np.concatenate((ps_points, ss_points))



import os

import numpy as np
import pyvista as pv

ON_CI = 'CI' in os.environ


def test_solution(tmpdir):
    from ntrfc.cascade_case.solution import GenericCascadeCase

    inletname = tmpdir / "fake_inlet.vtk"
    outletname = tmpdir / "fake_outlet.vtk"

    fake_inlet = pv.Plane(direction=(1, 0, 0))

    fake_inlet["u"] = np.array([1] * fake_inlet.number_of_cells)
    fake_inlet["v"] = np.array([0] * fake_inlet.number_of_cells)
    fake_inlet["w"] = np.array([0] * fake_inlet.number_of_cells)
    fake_inlet["rho"] = np.array([1] * fake_inlet.number_of_cells)
    fake_inlet["U"] = np.stack([fake_inlet["u"], fake_inlet["v"], fake_inlet["w"]]).T
    fake_inlet.save(inletname)
    fake_outlet = pv.Plane(direction=(-1, 0, 0))

    fake_outlet["u"] = np.array([1] * fake_outlet.number_of_cells)
    fake_outlet["v"] = np.array([0] * fake_outlet.number_of_cells)
    fake_outlet["w"] = np.array([0] * fake_outlet.number_of_cells)
    fake_outlet["rho"] = np.array([1] * fake_outlet.number_of_cells)
    fake_outlet["U"] = np.stack([fake_outlet["u"], fake_outlet["v"], fake_outlet["w"]]).T

    fake_outlet.save(outletname)
    case = GenericCascadeCase()
    case.read_meshes(inletname, "inlet")
    case.read_meshes(outletname, "outlet")
    case.compute_avdr_inout_massave()
    assert case.avdr == 1, "should be avdr==1"


def test_animations(tmpdir):
    from ntrfc.cascade_case.solution import GenericCascadeCase
    import pyvista as pv

    if ON_CI:
        pv.start_xvfb()

    noslices = 3
    test_slices = [pv.Plane() for i in range(noslices)]
    slices = []
    ts = []
    for idx, slice in enumerate(test_slices):
        slice["U"] = np.zeros(slice.number_of_cells)
        slice = slice.point_data_to_cell_data()
        fpath = f"{tmpdir}/{idx}/someface.vtk"
        os.mkdir(f"{tmpdir}/{idx}")
        slices.append(fpath)
        slice.save(fpath)
        ts.append(idx)

    test_solution = GenericCascadeCase()
    test_solution.sliceseries.add_sliceset(slices, "some", ts)
    test_solution.sliceseries.create_animation("some", "U", tmpdir, "U.gif")


def test_postprocessing():
    import pyvista as pv
    from ntrfc.cascade_case.solution import GenericCascadeCase
    from ntrfc.cascade_case.utils.domain_utils import DomainParameters
    from ntrfc.turbo.airfoil_generators.naca_airfoil_creator import naca

    xs, ys = naca("6510", 256)
    points = pv.PolyData(np.stack([xs, ys, np.zeros(len(xs))]).T)
    domainparas = DomainParameters()
    domainparas.generate_params_by_pointcloud(points)
    domainparas.profile_points["p"] = [1] * domainparas.profile_points.number_of_points

    inlet = pv.Plane(direction=(1, 0, 0))
    inlet["u"] = np.ones((inlet.number_of_cells))
    inlet["v"] = np.zeros((inlet.number_of_cells))
    inlet["w"] = np.zeros((inlet.number_of_cells))
    inlet["U"] = np.stack([inlet["u"], inlet["v"], inlet["w"]]).T
    inlet["rho"] = np.array([1] * inlet.number_of_cells)
    inlet["p"] = np.array([1] * inlet.number_of_cells)
    inlet.ctp()
    outlet = pv.Plane(direction=(-1, 0, 0))
    outlet["u"] = np.ones((outlet.number_of_cells))
    outlet["v"] = np.zeros((outlet.number_of_cells))
    outlet["w"] = np.zeros((outlet.number_of_cells))
    outlet["U"] = np.stack([inlet["u"], inlet["v"], inlet["w"]]).T
    outlet["rho"] = np.array([1] * outlet.number_of_cells)
    outlet["p"] = np.array([1] * outlet.number_of_cells)
    outlet.ctp()

    # Initialize PostProcessing object
    postprocessing = GenericCascadeCase()
    postprocessing.mesh_dict["inlet"] = inlet
    postprocessing.mesh_dict["outlet"] = outlet
    postprocessing.domainparams = domainparas

    # Test compute_avdr method
    postprocessing.compute_avdr_inout_massave()
    assert postprocessing.avdr == 1

    # Test blade_loading method
    assert postprocessing.blade_loading() == 0



import numpy as np
import pyvista as pv
from scipy.interpolate import splprep, splev
from scipy.spatial import KDTree
from scipy.spatial import Voronoi

from ntrfc.geometry.line import lines_from_points, polyline_from_points
from ntrfc.geometry.plane import inside_poly
from ntrfc.math.vectorcalc import findNearest, vecDir, compute_minmax_distance_in_pointcloud


def clean_sites(sites, boundary, tolerance_factor=3e-2):
    # Build the KDTree from the "boundary" point cloud.
    tree = KDTree(boundary)

    # Query the KDTree to find the nearest neighbors for each point in "point_cloud_a."

    distances, _ = tree.query(sites)
    deg_res = 2

    theta = np.linspace(0, 2 * np.pi, 360 * deg_res)
    cleaned = []
    radii = []
    for xc, yc, r in zip(sites[:, 0], sites[:, 1], distances):

        circlepoints = np.stack([np.cos(theta) * r + xc, np.sin(theta) * r + yc]).T
        circle_dists, _ = tree.query(circlepoints)

        half_circle_idx = 180 * deg_res
        circle_dists_180_deg = circle_dists[:half_circle_idx] + circle_dists[half_circle_idx:]
        # min_dist = np.argmin(circle_dists)
        atol = r * tolerance_factor
        circle_in_skeleton = np.where(np.isclose(circle_dists_180_deg, 0, atol=atol))[0]

        if len(circle_in_skeleton) > 0:
            min_dist_at = np.argmin(circle_dists[circle_in_skeleton])
            if np.isclose(circle_dists[circle_in_skeleton[min_dist_at]], 0, atol=atol / 2):
                cleaned.append([xc, yc])
                radii.append(r)
    print(f"ratio of cleaned sites: {len(cleaned) / len(sites)}")

    return np.array(cleaned), np.array(radii)


def extract_vk_hk(sortedPoly, verbose=False):
    points = sortedPoly.points

    points_2d_closed_refined = pointcloud_to_profile(points)

    sites_raw_clean, radii = voronoi_skeleton_sites(points_2d_closed_refined)

    tck, u = splprep(sites_raw_clean.T, u=None, s=0.000005, per=0, k=3)
    res = 4000
    u_new = np.linspace(u.min(), u.max(), res)
    x_center, y_center = splev(u_new, tck, der=0)
    le_ind, te_ind = skeletonline_completion(x_center, y_center, points)

    return le_ind, te_ind


def voronoi_skeleton_sites(points_2d_closed_refined):
    vor = Voronoi(points_2d_closed_refined)
    voronoi_sites_inside = vor.vertices[inside_poly(points_2d_closed_refined, vor.vertices)]

    sort_indices = np.argsort(voronoi_sites_inside[:, 0])
    sites_inside_sorted = voronoi_sites_inside[sort_indices]

    clean_sites_inside, radii = clean_sites(sites_inside_sorted, points_2d_closed_refined)
    return clean_sites_inside, radii


def skeletonline_completion(x_center, y_center, points):
    camberpoints = np.stack([x_center, y_center, np.zeros(len(y_center))]).T
    camberline = lines_from_points(camberpoints)
    LE_camber = camberline.extract_cells(0)
    LE_dir = vecDir(LE_camber.points[-1] - LE_camber.points[0])
    TE_camber = camberline.extract_cells(camberline.number_of_cells - 1)
    TE_dir = vecDir(TE_camber.points[0] - TE_camber.points[-1])
    X = points[::, 0]
    Y = points[::, 1]
    Z = X * 0
    profilepoly = polyline_from_points(np.vstack([np.stack([X, Y, Z]).T, np.stack([X[0], Y[0], Z[0]]).T]))
    chord = compute_minmax_distance_in_pointcloud(points, minmax="max")
    camber_le_extension = pv.Line(LE_camber.points[0] - LE_dir * chord / 2, camberpoints[0], resolution=400)
    camber_te_extension = pv.Line(camberpoints[-1], TE_camber.points[0] - TE_dir * chord / 2, resolution=400)
    camberline_extended = lines_from_points(np.vstack([camber_le_extension.points,
                                                       camberpoints[1:-2],
                                                       camber_te_extension.points]))
    helpersurface = profilepoly.copy().extrude([0, 0, -1], inplace=True)
    helpersurface = helpersurface.translate([0, 0, .5], inplace=True)
    camberline_computed = camberline_extended.clip_surface(helpersurface, invert=False)
    le_ind = findNearest(np.stack([X, Y, Z]).T, camberline_computed.points[0])
    te_ind = findNearest(np.stack([X, Y, Z]).T, camberline_computed.points[-1])

    return le_ind, te_ind


def pointcloud_to_profile(points):
    points_2d_closed = np.vstack((points[:, :2], points[:, :2][0]))
    tck, u = splprep(points_2d_closed.T, u=None, s=0.0, per=1, k=3)
    res = 10000
    u_new = np.linspace(u.min(), u.max(), res)
    x_new, y_new = splev(u_new, tck, der=0)
    points_2d_closed_refined = np.stack([x_new, y_new]).T
    return points_2d_closed_refined
